/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Utility functions.

****************************************************************************/

#include "all.h"

FILE * g_dbgFile;
struct WMSG {
	WMSG * next;
	char * msg;
};

static WMSG * s_messages;
static Bool s_buffering;
Seed g_defaultSeed = 1;

/////////////////////////////////////////////////////////////////////////////
//  Do we want buffering or don't we?
void WarnSetBuffering(Bool on) {
	s_buffering = on;
}

/////////////////////////////////////////////////////////////////////////////
//  This will flush the warning buffer.  I'm doing this so that we don't miss
// any messages that were present during a Lock.
void WarnBufferFlush(void) {
	WMSG * m;

	for (m=s_messages; m; ) {
		WMSG * next = m->next;

		if (m->msg) {
			fputs(m->msg, g_dbgFile);
			#if defined(WIN32)
				OutputDebugString(m->msg);
			#endif
			free(m->msg);
		}
		SafeDelete(m);
		m = next;
	}
	s_messages = NULL;
}

/////////////////////////////////////////////////////////////////////////////
//  Warn the user of a condition that has transpired.
void __cdecl Warn(const char * fmt, ...) {
	#if !defined(RELEASE_BUILD)
		va_list v;
		char buf[1024];
		WMSG * m;
	
		if (!g_dbgFile) {
			g_dbgFile = fopen("dbgout.txt", "w");
			setvbuf(g_dbgFile, NULL, _IONBF, 0);
		}
	
		va_start(v, fmt);
		vsprintf(buf, fmt, v);
		va_end(v);
	
		SafeNew(m, (MType)1, __FILE__, __LINE__);
		m->next = s_messages;
		m->msg = strdup(buf);
		s_messages = m;
		if (!s_buffering)
			WarnBufferFlush();
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Compute a random number in the indicated range.
Real Random(Real range, Seed & s) {
	s = 663608941 * s;
	return (Real)(range * (Real)s * (Real)(1.0 / (65536.0 * 65536.0)));
}
