/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Our datatypes, etc.

****************************************************************************/

#ifndef BASE_H
	#define BASE_H

	#if defined(WIN32)
		typedef unsigned char U8;
		typedef unsigned short U16;
		typedef unsigned int U32;
		typedef unsigned __int64 U64;
		typedef signed char S8;
		typedef signed short S16;
		typedef signed int S32;
		typedef signed __int64 S64;
	
		typedef U32 u8;
		typedef U32 u16;
		typedef U32 u32;
		typedef U64 u64;
		typedef S32 s8;
		typedef S32 s16;
		typedef S32 s32;
		typedef S64 s64;

		typedef double Real;
		typedef float real;

		typedef u32 MType;
		typedef int Bool;

		typedef U16 Color;
		typedef u16 color;

		#define BUFFER_COUNT				2

		extern HWND g_hWnd;
	#elif defined(DOS)
		typedef unsigned char U8;
		typedef unsigned short U16;
		typedef unsigned int U32;
		typedef signed char S8;
		typedef signed short S16;
		typedef signed int S32;
	
		typedef U32 u8;
		typedef U32 u16;
		typedef U32 u32;
		typedef S32 s8;
		typedef S32 s16;
		typedef S32 s32;

		typedef double Real;
		typedef float real;

		typedef u32 MType;
		typedef int Bool;

		typedef U16 Color;
		typedef u16 color;

		#define BUFFER_COUNT				2

	#endif

	#if !defined(TRUE)
		#define TRUE		1
	#endif

	#if !defined(FALSE)
		#define FALSE		0
	#endif

	#if !defined(NULL)
		#define NULL		0
	#endif

	extern void __cdecl Warn(const char * fmt, ...);
	extern void WarnSetBuffering(Bool on);
	extern void WarnBufferFlush(void);

	#if !defined(NDEBUG)
		#define CHECK(x) { if (!(x)) { Warn("CHECK failure, file %s, line %d.\n", __FILE__, __LINE__); } }
	#else
		#define CHECK(x)
	#endif

	// These should be relocated and optimized.
	inline real Recipsqrt(real x) { return 1.0 / sqrt(x); }
	inline real Reciprocoal(real x) { return 1.0 / x; }
	template < class A > inline A Abs(A a) { return (a < 0) ? -a : a; }
	template < class A > inline A Sign(A a) { if (a < 0) return -1; else if (a > 0) return 1; else return 0; }
	template < class A > inline void Swap(A & a, A & b) { A c; c=a; a=b; b=c; }
	template < class A > inline A Min(A a, A b) { if (a < b) return a; else return b; }
	template < class A > inline A Max(A a, A b) { if (a < b) return b; else return a; }
	inline real Epsilon(real r) { return 0.00001; }
	#if !defined(DOS)
		inline Real Epsilon(Real r) { return 1.0e-10; }
	#endif
	class IString;
	extern void Shutdown(const IString & i);

	struct Rect {
		u16 x;
		u16 y;
		u16 w;
		u16 h;
	};

	// These are values that have been derived from ini files, registry,
	// command line, etc.
	extern u16 g_userWidth, g_userHeight;	// User-preferred resolution.
	extern u8 g_userDepth;					// User-preferred color-depth.

	#if defined(DOS)
		template < class C >
		inline Bool operator ==(const C & a, const C & b) { return !memcmp(&a, &b, sizeof(C)); }
		template < class C >
		inline Bool operator !=(const C & a, const C & b) { return memcmp(&a, &b, sizeof(C)); }
	#endif

	typedef U32 Seed;

	extern Seed g_defaultSeed;

	extern void SendIPCMessage(u32 p1, u32 p2);
	Real Random(Real range, Seed & s = g_defaultSeed);
	extern "C" {
		extern Real __cdecl GetCPUSeconds(void);
	};
	typedef void (*AltEnterHandler)(void);
	extern AltEnterHandler g_altEnterHandler;

#endif
