/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of our big numbers.

****************************************************************************/

#include "all.h"

/////////////////////////////////////////////////////////////////////////////
//  Construct a bignum out of a u32.
BigNum::BigNum(u32 u) {
	m_rep.Make(2, __FILE__, __LINE__);
	m_rep[0] = u & 0xffff;
	m_rep[1] = u >> 16;
}

/////////////////////////////////////////////////////////////////////////////
//  Extract a bignum from a string.
BigNum::BigNum(const IString & i) {
	MType m;
	List < U16 > l;

	for (m=0; m<i.Length(); m+=4) {
		u8 a, b;
		U16 u=0;

		for (b=0; b<4; ++b) {
			if (i[m] >= 'A') {
				a = i[m + b] - 'A' + 10;
			} else {
				a = i[m + b] - '0';
			}
			u = (u << 4) + a;
		}
		l += u;
	}
	m_rep = l;
}

/////////////////////////////////////////////////////////////////////////////
//  Convert this bignum to an IString.
void BigNum::ConvertToString(IString & i) {
	MType m;

	for (m=0; m<m_rep.Count(); ++m) {
		U16 u = m_rep[m];
		u8 a, b;

		for (b=0; b<4; ++b) {
			a = (u >> 12) & 0xf;
			u <<= 4;

			if (a >= 0xa) {
				i += (wchar_t)('A' + a - 10);
			} else {
				i += (wchar_t)('0' + a);
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Add two bignums.
BigNum BigNum::operator +(const BigNum & c) const {
	u16 carry=0;
	List < U16 > l;
	MType m, n, x;
	BigNum retval;

	n = Min(c.m_rep.Count(), m_rep.Count());
	x = Max(c.m_rep.Count(), m_rep.Count());
	for (m=0; m<n; ++m) {
		u32 u = (u32)c.m_rep[m] + (u32)m_rep[m] + (u32)carry;

		l += u & 0xffff;
		carry = u >> 16;
	}
	for (;m<x; ++m) {
		if (x == m_rep.Count()) {
			l += (u32)m_rep[m] + (u32)carry;
		} else {
			l += (u32)c.m_rep[m] + (u32)carry;
		}
		carry = 0;
	}
	retval.m_rep = l;
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Shift c by the specified amount and add to b.
BigNum BigNum::ShiftAndAdd(const BigNum & c, MType shift) const {
	BigNum d;
	MType m;

	d.m_rep.Make(c.m_rep.Count() + shift, __FILE__, __LINE__);
	for (m=0; m<shift; ++m) {
		d.m_rep[m] = 0;
	}
	for (m=0; m<c.m_rep.Count(); ++m) {
		d.m_rep[m + shift] = c.m_rep[m];
	}
	return *this + d;
}

/////////////////////////////////////////////////////////////////////////////
//  Multiply a bignum times a U16.
BigNum BigNum::operator *(const U16 c) const {
	BigNum retval;
	List < U16 > l;
	MType m;
	u16 carry=0;

	for (m=0; m<m_rep.Count(); ++m) {
		u32 u = m_rep[m] * c + carry;

		l += u & 0xffff;
		carry = u >> 16;
	}
	if (carry) {
		l += carry;
	}
	retval.m_rep = l;
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Multiply two bignums.
BigNum BigNum::operator *(const BigNum & c) const {
	BigNum retval;
	MType m;

	if (m_rep.Count() < c.m_rep.Count()) {
		for (m=0; m<m_rep.Count(); ++m) {
			retval = retval.ShiftAndAdd(c * m_rep[m], m);
		}
	} else {
		for (m=0; m<c.m_rep.Count(); ++m) {
			retval = retval.ShiftAndAdd(*this * c.m_rep[m], m);
		}
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Determine the quotient and remainder when this number is divided by the
// indicated U16.
BigNum BigNum::QuotientRemainder(U16 & rem, const U16 c) const {
	u32 u, v, remainder=0;
	MType m;
	BigNum retval;
	List < U16 > l;

	for (m=m_rep.Count() - 1; TRUE; --m) {
		u = (remainder << 16) + m_rep[m];

		v = u / c;
		l += v;
		remainder = u - (u32)c * v;
		if (!m)			// Needed for MICROSUX.
			break;
	}
	l.Flip();
	while ((l.Count() > 1) && !l[l.Count() - 1]) {
		l.Chop();
	}
	retval.m_rep = l;
	rem = remainder;
	return retval;
}
