/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	All of our 2D color map rendering primitives.

****************************************************************************/

#ifndef BITMAP_H
	#define BITMAP_H

	/////////////////////////////////////////////////////////////////////////
	//	These bitmaps fill up their rectangles.
	class Bitmap : public Target < Color > {
		friend class Overlay;
		public:
			enum {
				CUTOUT		= 1,
				TRANSLUCENT	= 2,
				COLORIZE	= 4,
			};
			Bitmap(void);
			~Bitmap(void);
			void Render(Receptor & dest,
						u16 x,
						u16 y,
						RenderOptions & ro);
			void Import(const FString & name);
			void Export(const FString & name);
			void ExtractFrom(Map < Color > & m,
							 u16 x,
							 u16 y,
							 u16 w,
							 u16 h);
			Bitmap & operator =(const Bitmap & t);
			virtual Color MapColor(real, real, real) const;
			virtual void Clear(Color u, Rect * p = NULL);
		protected:
			void BuildRLE(color key);
			void MakeRLE(color key);
			virtual void MakeDirty(void);	// RLEs must be rebuilt.

			struct RLE {
				S16 jumpX;
				U16 jumpY;
				MType run;
			};
			struct RLEData {
				color key;
				Array < RLE > image;
				Array < Color > colorData;
			};
			RLEData * m_rle;
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	Overlays do not modify the view.  This can be done in hardware, or
	// emulated.
	class Overlay : public Bitmap {
		friend class Target < color >;
		public:
			void Render(Receptor & dest,
						u16 x,
						u16 y,
						RenderOptions & ro);
			void ActuallyDraw(Receptor & r,
							  u16 x,
							  u16 y,
							  RenderOptions & ro);
			void Erase(Receptor & r,
					   u16 x,
					   u16 y);
			void MakeDirty(void);
		protected:
		private:
			struct Save {
				Bitmap bitmap;
				u8 options;
				color key;
			};

			Vec < Save, BUFFER_COUNT > m_saveData;
	};

	typedef Object2D < Bitmap, RenderOptions > Sprite;
	typedef Object2D < Overlay, RenderOptions > Cursor;

	#include "bitmap.ipp"

#endif
