/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	This gives us some directory-browsing functions.

****************************************************************************/

#include "all.h"

#if defined(DOS)
	// Watcom doesn't provide these, so we will have to!
	inline wchar_t towupper(wchar_t t) {
		if ((t >= 'a') && (t <= 'z'))
			return t - 'a' + 'A';
		else
			return t;
	}
#endif

char g_appPath[256];

/////////////////////////////////////////////////////////////////////////////
//  This extracts the filename from the FString.
MType DirFormFilename(char * buf, const FString & s) {
	MType m, n;

	for (m=n=0; m<s.Length(); ++m) {
		if (s[m] >= TEXT_NUM_CONTROL_CODES)
			buf[n++] = s[m];
	}
	buf[n] = '\0';
	return n;
}

/////////////////////////////////////////////////////////////////////////////
//  Get a list of all files here.
Bool DirFindAll(List < IString > & dest, const FString & fn) {
	#if defined(WIN32)
		long fileHandle;
		struct _finddata_t f;
		char buf[256];
		MType n;

CopyToChar(fn, buf, 256);
Warn("Doing FindAll:  %s.\n", buf);

		dest.Destroy();
		// First, let's list all drives.
		buf[0] = 'A';
		buf[1] = ':';
		buf[2] = '/';
		buf[3] = '\0';
		for (n=0; n<32; ++n) {
			buf[0] = 'A' + n;
			switch(GetDriveType(buf)) {
				case DRIVE_UNKNOWN:
				case DRIVE_NO_ROOT_DIR:
					break;
				default: {
					IString & s = dest.Grow();

					s = DIR_DRIVE;
					s += buf;
					break;
				}
			}
		}

		n = DirFormFilename(buf, fn);
		if (buf[n-1] != '/')
			buf[n++] = '/';
		buf[n++] = '*';
		buf[n++] = '.';
		buf[n++] = '*';
		buf[n] = '\0';
		if ((fileHandle = _findfirst(buf, &f)) != -1) {
			do {
				IString & s = dest.Grow();
	
				if (f.attrib & _A_SUBDIR) {
					s = DIR_DIRECTORY;
				} else {
					s = DIR_FILE;
				}
				s += f.name;
				for (n=0; n<s.Length(); ++n) {
					s[n] = towupper(s[n]);
				}
			} while (!_findnext(fileHandle, &f));
			_findclose(fileHandle);
			Sort(dest);
			return TRUE;
		} else {
			return FALSE;
		}
	#else
		struct find_t f;
		char buf[256];
		MType n;
	
		n = DirFormFilename(buf, fn);
		if (buf[n-1] != '/')
			buf[n++] = '/';
		buf[n++] = '*';
		buf[n++] = '.';
		buf[n++] = '*';
		buf[n] = '\0';
		dest.Destroy();
		if (_dos_findfirst(buf, _A_NORMAL | _A_SUBDIR, &f) == 0) {
			do {
				IString & s = dest.Grow();

				if (f.attrib & _A_SUBDIR) {
					s = DIR_DIRECTORY;
				} else {
					s = DIR_FILE;
				}
				s += f.name;
				for (n=0; n<s.Length(); ++n) {
					s[n] = towupper(s[n]);
				}
			} while (_dos_findnext(&f) == 0);
			_dos_findclose(&f);
			Sort(dest);
			return TRUE;
		} else {
			return FALSE;
		}
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Get the current working directory.
FString DirGetCurrent(void) {
	char buf[256];
	FString retval = DIR_DIRECTORY;

	#if defined(WIN32)
		retval += _getcwd(buf, 256);
	#elif defined(DOS)
		retval += getcwd(buf, 256);
	#endif
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Change the current directory to the one indicated.
void DirChange(const FString & dest) {
	char buf[256];

	DirFormFilename(buf, dest);
	#if defined(WIN32)
		_chdir(buf);
	#elif defined(DOS)
		chdir(buf);
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Read in a (binary) file.
Array < U8 > DirLoadFile(const FString & filename) {
	char buf[256];
	FILE * f;
	List < U8 > retval;
	U8 lclBuf[1024];
	int i;
	MType m;

	DirFormFilename(buf, filename);
	if (!(f = fopen(buf, "rb"))) {
		Warn("Unable to open file %s.\n", buf);
	} else {
		do {
			i = fread(lclBuf, 1, 1024, f);
			for (m=0; m<i; ++m) {
				retval += lclBuf[m];
			}
		}  while (i == 1024);
		fclose(f);
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Append the indicated directory or filename to the indicated directory or
// filename.
void DirAppend(FString & s, const FString & d) {
	FString c;
	MType m;
	u8 n;

	for (m=0; m<d.Length(); ++m) {
		if (d[m] >= TEXT_NUM_CONTROL_CODES)
			c += d[m];
	}

	if (c.Length() == 0)
		return;
	if (c[c.Length() - 1] == '/') {
		c.Chop();
	}
	if (c.Length() == 1) {
		if (c[0] == '.') {
			return;
		}
	}
	if (s[s.Length() - 1] != '/') {
		s += '/';
	}
	if (c.Length() == 2) {
		if ((c[0] == '.') && (c[1] == '.')) {
			for (m=s.Length(), n=2; n && m--; ) {
				if (s[m] == '/') {
					--n;
				}
			}
			// Never found that second slash.
			if (n) {
				s = "";
				return;
			} else {
				s.Chop(s.Length() - m);
				return;
			}
		}
	}

	// If we come here, it's a normal appendage.
	s += c;
}

/////////////////////////////////////////////////////////////////////////////
//  Get the path that the application was launched from.
FString DirGetAppPath(void) {
	return g_appPath;
}

/////////////////////////////////////////////////////////////////////////////
//  Does the indicated directory exist?
Bool DirExists(const FString & f) {
	long fileHandle;
	struct _finddata_t fd;
	char buf[MAX_PATH + 1];
	MType m;
	Bool retval;

	CopyToChar(f, buf, MAX_PATH + 1);
	m = strlen(buf);
	if (buf[m-1] != '/')
		buf[m++] = '/';
	buf[m++] = '*';
	buf[m++] = '.';
	buf[m++] = '*';
	buf[m++] = '\0';
	if ((fileHandle = _findfirst(buf, &fd)) != -1) {
		retval = TRUE;
		_findclose(fileHandle);
	} else {
		retval = FALSE;
	}
	return retval;
}
