/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Strings with international support.

****************************************************************************/

#ifndef ISTRING_H
	#define ISTRING_H

	#define FSTRING_SIZE				256

	class FString;
	class TString;

	class IString : public List < wchar_t > {
		public:
			void DebugOut(void) const;
			IString(const wchar_t c);
			IString(const char * s);
			IString(void);

			IString & operator =(const char * s);
			IString & operator =(const FString & f);
			IString & operator =(const wchar_t c);

			IString & operator +=(const wchar_t w);
			IString & operator +=(const char c);
			IString & operator +=(const IString & i);
			IString & operator +=(const char * s);

			Bool operator <(const IString & i) const;
			Bool operator !=(const IString & i) const;

			MType Length(void) const;
		protected:
		private:
	};

	class TString : public IString {
		public:
			TString(const char * s);
		protected:
		private:
	};

	class FString : public Vec < wchar_t, FSTRING_SIZE > {
		public:
			void DebugOut(void) const;
			void Destroy(void);

			FString(const IString & i);
			FString(const wchar_t c);
			FString(const char * s);
			FString(void);

			FString & operator =(const wchar_t c);
			FString & operator =(const char * s);
			FString & operator =(const FString & f);
			FString & operator =(const IString & i);

			FString & operator +=(const wchar_t w);
			FString & operator +=(const char c);
			FString & operator +=(const FString & i);
			FString & operator +=(const char * s);
			FString & operator +=(const IString & i);
			FString & operator +=(const wchar_t * s);

			Bool operator <(const FString & i) const;
			Bool operator !=(const FString & i) const;

			void Chop(MType nChars=1);

			MType Length(void) const;
		protected:
		private:
			MType m_cursor;
	};

	/////////////////////////////////////////////////////////////////////////
	//	These are generic operations that can be performed on any string.
	template < class S > u32 ConvertToHex(S & s);
	template < class S > void ConvertFromHex(S & s, u32 v, u8 numDigits);
	template < class S > void ConvertFromHex(S & s, u32 v);
	template < class S > void ConvertFromDecimal(S & s, u32 v);
	template < class S > void ConvertFromFloat(S & s, real r, u8 numDigits);
	template < class S > void CopyToChar(const S & s, char * b, MType l);

	#include "istring.ipp"

	extern List < FString > g_args;

#endif
