/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The inlined implementation of internationalized strings.

****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
//  Made the default constructor to allow for the non-default.
inline IString::IString(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  Append a wchar to an IString.
inline IString & IString::operator +=(const wchar_t w) {
	List < wchar_t >::operator +=(w);
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append an ascii char to an IString.
inline IString & IString::operator +=(const char c) {
	wchar_t w = (wchar_t)c;

	List < wchar_t >::operator +=(w);
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append one istring to another.
inline IString & IString::operator +=(const IString & i) {
	List < wchar_t >::operator +=((List < wchar_t >)i);
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append an ascii string to this puppy.
inline IString & IString::operator +=(const char * s) {
	MType slen = strlen(s);
	MType n = m_numberOfElements + slen;

	if (n >= m_nextAlloc) {
		wchar_t * rep;
		MType m;

		while (m_nextAlloc < n) {
			m_nextAlloc += m_nextAlloc;
		}
		SafeNew(rep, m_nextAlloc, __FILE__, __LINE__);
		for (m=0; m<m_numberOfElements; ++m) {
			rep[m] = m_elements[m];
		}
		SafeDelete(m_elements);
		m_elements = rep;
	}
	for (n=0; n<slen; ++n) {
		m_elements[n + m_numberOfElements] = (wchar_t)s[n];
	}
	m_numberOfElements += slen;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Get the length of the IString.
inline MType IString::Length(void) const {
	return Count();
}

/////////////////////////////////////////////////////////////////////////////
//  Destroy the contents of the FString.
inline void FString::Destroy(void) {
	m_cursor = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Make an FString out of a wide character.
inline FString::FString(const wchar_t c) {
	m_elements[0] = c;
	m_cursor = 1;
}

/////////////////////////////////////////////////////////////////////////////
//  Make an FString out of a C string.
inline FString::FString(const char * s) {
	m_cursor = 0;
	while (*s) {
		CHECK(m_cursor < FSTRING_SIZE);
		m_elements[m_cursor] = *s++;
		++m_cursor;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Make an FString out of an IString.
inline FString::FString(const IString & i) {
	for (m_cursor=0; m_cursor<i.Length(); ++m_cursor) {
		CHECK(m_cursor < FSTRING_SIZE);
		m_elements[m_cursor] = i[m_cursor];
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Make any old FString.
inline FString::FString(void) {
	m_cursor = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Assign a C string to an FString.
inline FString & FString::operator =(const char * s) {
	m_cursor = 0;
	while (*s) {
		if (m_cursor == FSTRING_SIZE)
			break;
		m_elements[m_cursor] = *s++;
		++m_cursor;
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Assign a FString to an FString.
inline FString & FString::operator =(const FString & f) {
	for (m_cursor=0; m_cursor < f.Length(); ++m_cursor) {
		m_elements[m_cursor] = f[m_cursor];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Assign an IString to an FString.
inline FString & FString::operator =(const IString & i) {
	for (m_cursor=0; m_cursor < i.Length(); ++m_cursor) {
		if (m_cursor == FSTRING_SIZE) {
			break;
		}
		m_elements[m_cursor] = i[m_cursor];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Assign a wchar_t to an FString.
inline FString & FString::operator =(const wchar_t w) {
	m_elements[0] = w;
	m_cursor = 1;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append a wchar_t.
inline FString & FString::operator +=(const wchar_t w) {
	CHECK(m_cursor < FSTRING_SIZE);
	m_elements[m_cursor++] = w;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append a char.
inline FString & FString::operator +=(const char c) {
	CHECK(m_cursor < FSTRING_SIZE);
	m_elements[m_cursor++] = c;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append another FString.
inline FString & FString::operator +=(const FString & i) {
	MType m;

	for (m=0; m<i.Length(); ++m) {
		CHECK(m_cursor < FSTRING_SIZE);
		m_elements[m_cursor] = i[m];
		++m_cursor;
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append a C string.
inline FString & FString::operator +=(const char * s) {
	while (*s) {
		CHECK(m_cursor < FSTRING_SIZE);
		m_elements[m_cursor] = *s++;
		++m_cursor;
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append a wide-character C string.
inline FString & FString::operator +=(const wchar_t * s) {
	while (*s) {
		CHECK(m_cursor < FSTRING_SIZE);
		m_elements[m_cursor] = *s++;
		++m_cursor;
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Append an IString.
inline FString & FString::operator +=(const IString & i) {
	MType m;

	for (m=0; m<i.Length(); ++m) {
		CHECK(m_cursor < FSTRING_SIZE);
		m_elements[m_cursor] = i[m];
		++m_cursor;
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  How long is the string.
inline MType FString::Length(void) const {
	return m_cursor;
}

/////////////////////////////////////////////////////////////////////////////
//  The string is assumed to contain a hexadecimal value.  We do a greedy
// conversion, terminating only when we reach a non-hex digit.
template < class S >
u32 ConvertToHex(S & s) {
	MType m;
	u32 retval=0;

	for (m=0; m<s.Length(); ++m) {
		if (s[m] < 32) {
			continue;
		}
		if ((s[m] >= '0') && (s[m] <= '9')) {
			retval = (retval << 4) | (s[m] - '0');
		} else if ((s[m] >= 'A') && (s[m] <= 'F')) {
			retval = (retval << 4) | ((s[m] - 'A') + 0xa);
		} else if ((s[m] >= 'a') && (s[m] <= 'f')) {
			retval = (retval << 4) | ((s[m] - 'a') + 0xa);
		} else {
			return retval;
		}
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  This converts a hex constant to an n-character string.
template < class S >
void ConvertFromHex(S & s, u32 v, u8 numDigits) {
	u8 u;

	for (u=0; u<8 - numDigits; ++u) {
		v <<= 4;
	}
	for (u=0; u<numDigits; ++u) {
		u8 w = (v >> 28) & 0xf;

		if (w > 9) {
			s += (wchar_t)(w - 0xa + 'A');
		} else {
			s += (wchar_t)(w + '0');
		}
		v <<= 4;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This converts from a decimal constant to an n-character string.
template < class S >
void ConvertFromDecimal(S & s, u32 v) {
	u8 u;
	char buf[256];

	if (!v) {
		s += (wchar_t)'0';
	} else {
		buf[255] = '\0';
		for (u=0; v; ++u) {
			buf[254 - u] = (v % 10) + '0';
			v /= 10;
		}
		s += (buf + 255 - u);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  MICROSOFT SUCKS!
template < class S >
void ConvertFromHex(S & s, u32 v) {
	ConvertFromHex(s, v, 8);
}

/////////////////////////////////////////////////////////////////////////////
//  Convert a float to a string.
template < class S >
void ConvertFromFloat(S & s, real r, u8 numDigits) {
	char buf[256];
	u8 c;

	sprintf(buf, "%lg", (double)r);
	for (c=0; buf[c] && (c<numDigits); ++c) {
		s += buf[c];
	}
	while (c < numDigits) {
		s += ' ';
		++c;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Transfer data from one of my strings to an ascii buffer.
template < class S > void CopyToChar(const S & s, char * b, MType l) {
	MType m;

	if (l > s.Length())
		l = s.Length();
	else
		--l;
	for (m=0; m<l; ++m) {
		b[m] = s[m];
	}
	b[m] = '\0';
}
