/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Specializations of various template functions.

****************************************************************************/

#include "all.h"

Calculator < JM_LOW_PRECISION > LCalculator;
Calculator < JM_MED_PRECISION > MCalculator;
// Calculator < JM_HIGH_PRECISION > HCalculator;
vec3 g_workVecs[4];
real g_cosTable[1024];

/////////////////////////////////////////////////////////////////////////////
//  Build the cosine table.
static void BuildCosineTable(void) {
	MType m;

	for (m=0; m<1024; ++m) {
		Real r;

		r = m * 2.0 * JM_PI / 1024.0;
		g_cosTable[m] = (real)cos(r);
	}
}

static Initializer s_BuildCosineTable(BuildCosineTable);

/////////////////////////////////////////////////////////////////////////////
//  Compute the cross product with another vector, store in this one.  This
// implementation only applies to 3-dimensional vectors of reals (MSVC does
// not seem to support partial specialization)
void
MathVec < Real, 3 >::CrossProduct(MathVec < Real, 3 > & w) {
	Elem a, b;

	a = m_elements[1] * w.m_elements[2] - m_elements[2] * w.m_elements[1];
	b = m_elements[0] * w.m_elements[2] - m_elements[2] * w.m_elements[0];
	m_elements[2] = m_elements[0] * w.m_elements[1] -
					m_elements[1] * w.m_elements[0];
	m_elements[0] = a;
	m_elements[1] = b;
}

/////////////////////////////////////////////////////////////////////////////
//  Three-operand cross product.
void
MathVec < Real, 3 >::CrossProduct(MathVec < Real, 3 > & w,
								  MathVec < Real, 3 > & x) {
	m_elements[0] = w.m_elements[1] * x.m_elements[2] -
					w.m_elements[2] * x.m_elements[1];
	m_elements[1] = w.m_elements[0] * x.m_elements[2] -
					w.m_elements[2] * x.m_elements[0];
	m_elements[2] = w.m_elements[0] * x.m_elements[1] -
					w.m_elements[1] * x.m_elements[0];
}

