/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Inevitably, we need certain classes for math functions and objects.

****************************************************************************/

#ifndef JMATH_H
	#define JMATH_H

/****************************************************************************
*****************************************************************************

						S	C	A	L	A	R	S

*****************************************************************************
****************************************************************************/

	template < class R >
	R Sin(R r) { return (R)sin((double)r); }

/****************************************************************************
*****************************************************************************

						V	E	C	T	O	R	S

*****************************************************************************
****************************************************************************/


	template < class Elem, MType t_dim >
	class MathVec : public Vec < Elem, t_dim > {
		public:
			MathVec & operator =(Elem e);
			MathVec & operator +=(const MathVec & w);
			void CrossProduct(MathVec & w);
			void CrossProduct(MathVec & w, MathVec & x);
			void Negate(void);
			Elem Unit(void);

			// Don't use this function.  I wanted to hide it, but then I
			// would have to list too many friends.
			Elem * CastToFloatPtr(void);
		protected:
		private:
	};

	// Binary ops are non-members.
	template < class Elem, MType t_dim >
	MathVec < Elem, t_dim > operator -(const MathVec < Elem, t_dim > & v,
									   const MathVec < Elem, t_dim > & w);
	template < class Elem, MType t_dim >
	MathVec < Elem, t_dim > operator *(const MathVec < Elem, t_dim > & v,
									   Elem r);

	// Dot product.
	template < class Elem, MType t_dim >
	Elem operator *(const MathVec < Elem, t_dim > & v,
					const MathVec < Elem, t_dim > & w);

	template < class Elem, MType t_dim >
	class SqrMatrix : public Vec < MathVec < Elem, t_dim >, t_dim > {
		public:
			Bool Invert(void);
		protected:
		private:
	};

	typedef MathVec < Real, 3 > Vec3;
	typedef MathVec < real, 3 > vec3;

	// Use these wherever possible, for cache hits.  Be careful!  They're
	// neither recursion-safe nor fcall safe.
	extern vec3 g_workVecs[4];

/****************************************************************************
*****************************************************************************

					F	A	S	T		M	A	T	H

*****************************************************************************
****************************************************************************/

	enum {
		JM_LOW_PRECISION,
		JM_MED_PRECISION,
		JM_HIGH_PRECISION,
		JM_DEGREES				= 65536,
	};
	#define JM_PI				3.141592654

	/////////////////////////////////////////////////////////////////////////
	//	All trig functions assume radians.  HEREHEREHERE we will want to do
	// some partial inlining here for performance, and we'll want to define
	// our fast-cast, fast-recip, sqrt, etc.
	template < u8 precision >
	class Calculator {
		public:
			static u16 RadToJDeg(real r);
			static real Reciprocoal(real r);
			static real JCos(u16 degs);
			static real JSin(u16 degs);
			static real Cos(real degs);
			static real Sin(real degs);
	};

	extern Calculator < JM_LOW_PRECISION > LCalculator;
	extern Calculator < JM_MED_PRECISION > MCalculator;
	// extern Calculator < JM_HIGH_PRECISION > HCalculator;

	extern real g_cosTable[];

	#include "jmath.ipp"

#endif
