/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	This contains our interface to the keyboard handler.
	Here's what we want to do:  we want to maintain a queue of keypresses
that we can delete from, AND we want to be able to poll all keys that were
down from the last time we checked to the present.

****************************************************************************/

#ifndef _KB_H_
	#define _KB_H_

	#if defined(WIN32)
		enum {
			KB_SHIFT		= 0x8000,
			KB_ALT			= 0x4000,
			KB_CTRL			= 0x2000,
	
			KB_BASE_MASK	= 0x0fff,
	
			KB_PGUP			= VK_PRIOR,
			KB_PGDN			= VK_NEXT,
			KB_HOME			= VK_HOME,
			KB_END			= VK_END,
			KB_INS			= VK_INSERT,
			KB_DEL			= VK_DELETE,
			KB_ENTER		= VK_RETURN,
			KB_SPACE		= VK_SPACE,
			KB_UP			= VK_UP,
			KB_DOWN			= VK_DOWN,
			KB_LEFT			= VK_LEFT,
			KB_RIGHT		= VK_RIGHT,
			KB_ESC			= VK_ESCAPE,
			KB_BACK			= VK_BACK,
			KB_F1			= VK_F1,
			KB_F2			= VK_F2,
			KB_F3			= VK_F3,
			KB_F4			= VK_F4,
			KB_F5			= VK_F5,
			KB_F6			= VK_F6,
			KB_F7			= VK_F7,
			KB_F8			= VK_F8,
			KB_F9			= VK_F9,
			KB_F10			= VK_F10,
			KB_PRTSCR		= VK_SNAPSHOT,
	
			KB_0			= '0',
			KB_1			= '1',
			KB_2			= '2',
			KB_3			= '3',
			KB_4			= '4',
			KB_5			= '5',
			KB_6			= '6',
			KB_7			= '7',
			KB_8			= '8',
			KB_9			= '9',
	
			KB_A			= 'A',
			KB_B			= 'B',
			KB_C			= 'C',
			KB_D			= 'D',
			KB_E			= 'E',
			KB_F			= 'F',
			KB_G			= 'G',
			KB_H			= 'H',
			KB_I			= 'I',
			KB_J			= 'J',
			KB_K			= 'K',
			KB_L			= 'L',
			KB_M			= 'M',
			KB_N			= 'N',
			KB_O			= 'O',
			KB_P			= 'P',
			KB_Q			= 'Q',
			KB_R			= 'R',
			KB_S			= 'S',
			KB_T			= 'T',
			KB_U			= 'U',
			KB_V			= 'V',
			KB_W			= 'W',
			KB_X			= 'X',
			KB_Y			= 'Y',
			KB_Z			= 'Z',
		};
	#elif defined(DOS)
		enum {
			KB_SHIFT			 = 0x8000,
			KB_ALT				 = 0x4000,
			KB_CTRL				 = 0x2000,
		
			KB_BASE_MASK		 = 0x0fff,
		
			KB_ESC				 = 0x1,
			KB_F1				 = 0x3b,
			KB_F2				 = 0x3c,
			KB_F3				 = 0x3d,
			KB_F4				 = 0x3e,
			KB_F5				 = 0x3f,
			KB_F6				 = 0x40,
			KB_F7				 = 0x41,
			KB_F8				 = 0x42,
			KB_F9				 = 0x43,
			KB_F10				 = 0x44,
			KB_F11				 = 0x85,
			KB_F12				 = 0x86,
			KB_NK_SLASH			 = 0xe0,
			KB_NK_ASTERISK		 = 0x37,
			KB_NK_MINUS			 = 0x4a,
			KB_BACK_APOSTROPHE	 = 0x29,
			KB_1				 = 0x2,
			KB_2				 = 0x3,
			KB_3				 = 0x4,
			KB_4				 = 0x5,
			KB_5				 = 0x6,
			KB_6				 = 0x7,
			KB_7				 = 0x8,
			KB_8				 = 0x9,
			KB_9				 = 0xa,
			KB_0				 = 0xb,
			KB_MINUS			 = 0xc,
			KB_EQUALS			 = 0xd,
			KB_BACK				 = 0xe,
			KB_INS				 = 0x52,
			KB_HOME				 = 0x47,
			KB_PGUP				 = 0x49,
			KB_NK_7				 = 0x47,
			KB_NK_8				 = 0x48,
			KB_NK_9				 = 0x49,
			KB_NK_PLUS			 = 0x4e,
			KB_TAB				 = 0xf,
			KB_Q				 = 0x10,
			KB_W				 = 0x11,
			KB_E				 = 0x12,
			KB_R				 = 0x13,
			KB_T				 = 0x14,
			KB_Y				 = 0x15,
			KB_U				 = 0x16,
			KB_I				 = 0x17,
			KB_O				 = 0x18,
			KB_P				 = 0x19,
			KB_LBRACKET			 = 0x1a,
			KB_RBRACKET			 = 0x1b,
			KB_BACKSLASH		 = 0x2b,
			KB_DEL				 = 0x53,
			KB_END				 = 0x4f,
			KB_PGDN				 = 0x51,
			KB_NK_4				 = 0x4b,
			KB_NK_5				 = 0x4c,
			KB_NK_6				 = 0x4d,
			KB_A				 = 0x1e,
			KB_S				 = 0x1f,
			KB_D				 = 0x20,
			KB_F				 = 0x21,
			KB_G				 = 0x22,
			KB_H				 = 0x23,
			KB_J				 = 0x24,
			KB_K				 = 0x25,
			KB_L				 = 0x26,
			KB_SEMI				 = 0x27,
			KB_APOSTROPHE		 = 0x28,
			KB_ENTER			 = 0x1c,
			KB_NK_1				 = 0x4f,
			KB_NK_2				 = 0x50,
			KB_NK_3				 = 0x51,
			KB_NK_ENTER			 = 0xe0,
			KB_Z				 = 0x2c,
			KB_X				 = 0x2d,
			KB_C				 = 0x2e,
			KB_V				 = 0x2f,
			KB_B				 = 0x30,
			KB_N				 = 0x31,
			KB_M				 = 0x32,
			KB_COMMA			 = 0x33,
			KB_PERIOD			 = 0x34,
			KB_SLASH			 = 0x35,
			KB_UP				 = 0x48,
			KB_NK_0				 = 0x52,
			KB_NK_PERIOD		 = 0x53,
			KB_SPACE			 = 0x39,
			KB_LEFT				 = 0x4b,
			KB_DOWN				 = 0x50,
			KB_RIGHT			 = 0x4d,
		};
	#endif

	typedef u16 Key;

	Key KeyPop(void);
	Bool KeysThere(void);
	Bool KeyDown(Key k);

	#define KeyToChar(x) ((wchar_t)(x))

#endif
