/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Okay, this sets up the graphics engine, sound engine, thread engine, etc.
and calls the app's entry point.

****************************************************************************/

#include "all.h"

//  These must be defined by the application.
extern u8 MainThread(void *);
extern void InitializeApp(void);
extern void ShutdownApp(void);
extern Bool ProcessIPCMessage(u16 w, u32 d);
extern char * g_appName;
extern char * g_appTitle;

u16 g_ddrawWidth, g_ddrawHeight;
MType g_ddrawPitch;
u8 g_ddrawDepth;
color g_redShift, g_greenShift, g_blueShift;
u8 g_redMask, g_greenMask, g_blueMask;
u16 g_mouseX, g_mouseY;
u16 g_userWidth, g_userHeight;
u8 g_userDepth;
ThreadId g_mainThread, g_garbageThread, g_messagePumpThread;
extern Bool g_fullscreen;

extern RECT g_rcWindow;
extern RECT g_rcViewport;
extern RECT g_rcScreen;

HWND g_hWnd;
UINT g_IPCMessage;

extern FILE * g_dbgFile;

extern Key g_keyBuffer[];
extern U8 g_keyFront, g_keyBack;

HINSTANCE g_hInst;

LPDIRECTDRAW7			  	g_pDD = NULL;			// DirectDraw object
LPDIRECTDRAWSURFACE7		g_pDDSPrimary = NULL;	// DirectDraw primary surface
LPDIRECTDRAWSURFACE7		g_pDDSBack = NULL;  	// DirectDraw back surface
extern Bool g_active;
extern Bool g_ready;

View g_monitor;

extern Key g_keyMask;
AltEnterHandler g_altEnterHandler;

/////////////////////////////////////////////////////////////////////////////
//  Cleanup.
static void ReleaseAllObjects(void) {
	#if !defined(DUMMY_BUFFER)
	#if 0
		if (g_pDD != NULL) {
			if (g_pDDSPrimary != NULL) {
				g_pDDSPrimary->Release();
				g_pDDSPrimary = NULL;
			}
			g_pDD->Release();
			g_pDD = NULL;
		}
	#else
	    if (g_pDD != NULL) {
	        g_pDD->SetCooperativeLevel(g_hWnd, DDSCL_NORMAL);
	        if (g_pDDSBack != NULL) {
	            g_pDDSBack->Release();
	            g_pDDSBack = NULL;
	        }
	        if (g_pDDSPrimary != NULL) {
	            g_pDDSPrimary->Release();
	            g_pDDSPrimary = NULL;
	        }
			g_pDD->Release();
			g_pDD = NULL;
	    }
	#endif
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Cleanup between mode switches.
void ReleaseMostObjects(HWND hWnd) {
	#if !defined(DUMMY_BUFFER)
	    if (g_pDD != NULL) {
	        g_pDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL);
	        if (g_pDDSBack != NULL) {
	            g_pDDSBack->Release();
	            g_pDDSBack = NULL;
	        }
	        if (g_pDDSPrimary != NULL) {
	            g_pDDSPrimary->Release();
	            g_pDDSPrimary = NULL;
	        }
	    }
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  I should unify this function with the Warn function that already exists.
HRESULT __cdecl InitFail(HWND hWnd,
						 HRESULT hRet,
						 LPCTSTR szError,
						 ...) {
	char szBuff[128];
	va_list vl;

	va_start(vl, szError);
	vsprintf(szBuff, szError, vl);
	ReleaseAllObjects();
	MessageBox(hWnd, szBuff, g_appTitle, MB_OK);
	DestroyWindow(hWnd);
	va_end(vl);
	return hRet;
}

/////////////////////////////////////////////////////////////////////////////
//  A system that uses my strings.
void Shutdown(const IString & i) {
	char buf[256];

	CopyToChar(i, buf, 256);
	InitFail(g_hWnd, 0, buf);
}

/////////////////////////////////////////////////////////////////////////////
//  The message handler.
long FAR PASCAL WindowProc(HWND hWnd,
						   UINT message,
						   WPARAM wParam,
						   LPARAM lParam) {
	switch (message) {
		case WM_ACTIVATE:
			// Pause if minimized
			g_active = !((BOOL)HIWORD(wParam));
			return 0L;

		case WM_DESTROY:
			// Clean up and close the app
			ReleaseAllObjects();
			PostQuitMessage(0);
			return 0L;
		case WM_SYSKEYDOWN:
			if (wParam == 'Q') {
				PostMessage(hWnd, WM_CLOSE, 0, 0);
				return 0L;
			} else if (wParam == VK_RETURN) {
				if (g_altEnterHandler) {
					g_altEnterHandler();
				}
				return 0L;
			}
			break;
        case WM_PAINT:
            // Update the screen if we need to refresh
            if (g_ready && !g_fullscreen) {
                while (TRUE) {
					HRESULT hRet;

                    // If we are in windowed mode, perform a blt.
                    hRet = g_pDDSPrimary->Blt(&g_rcScreen,
                    						  g_pDDSBack,
                                              &g_rcViewport,
                                              DDBLT_WAIT,
                                              NULL);
                    if (hRet == DD_OK)
                        break;
                    if (hRet == DDERR_SURFACELOST) {
                        hRet = g_pDDSPrimary->Restore();
                        if (hRet != DD_OK )
                        	break;
                    }
                    if (hRet != DDERR_WASSTILLDRAWING)
                        break;
                }
            }
            break;
		case WM_MOVE:
            // Retrieve the window position after a move
            if (g_active && g_ready && !g_fullscreen) {
                GetWindowRect(hWnd, &g_rcWindow);
            	GetClientRect(hWnd, &g_rcViewport);
            	GetClientRect(hWnd, &g_rcScreen);
            	ClientToScreen(hWnd, (POINT*)&g_rcScreen.left);
            	ClientToScreen(hWnd, (POINT*)&g_rcScreen.right);
            }
            break;
		case WM_KEYDOWN:
			// I know I'm doing something wrong here, but I'll resolve that
			// later.
			switch (wParam) {
				case VK_SHIFT:
					break;
				case VK_CONTROL:
					break;
				case VK_MENU:
					break;
				default:
					CHECK(wParam <= 0xff);
					lParam &= 0xffff;
					while (lParam--)
						g_keyBuffer[g_keyBack++] = wParam;
					break;
			}
			break;
		case WM_LBUTTONDOWN:
			g_mouseX = LOWORD(lParam);
			if (g_mouseX >= g_ddrawWidth) {
				g_mouseX = g_ddrawWidth - 1;
			}
			g_mouseY = HIWORD(lParam);
			if (g_mouseY >= g_ddrawHeight) {
				g_mouseY = g_ddrawHeight - 1;
			}
			if (g_mouseDisplay) {
				MouseHotspotGroup::Click(g_mouseX, g_mouseY);
			}
			break;
		case WM_RBUTTONDOWN:
			g_mouseX = LOWORD(lParam);
			if (g_mouseX >= g_ddrawWidth) {
				g_mouseX = g_ddrawWidth - 1;
			}
			g_mouseY = HIWORD(lParam);
			if (g_mouseY >= g_ddrawHeight) {
				g_mouseY = g_ddrawHeight - 1;
			}
			if (g_mouseDisplay) {
				MouseHotspotGroup::RClick(g_mouseX, g_mouseY);
			}
			break;
		case WM_MOUSEMOVE:
			g_mouseX = LOWORD(lParam);
			if (g_mouseX >= g_ddrawWidth) {
				g_mouseX = g_ddrawWidth - 1;
			}
			g_mouseY = HIWORD(lParam);
			if (g_mouseY >= g_ddrawHeight) {
				g_mouseY = g_ddrawHeight - 1;
			}
			break;
		case WM_SETCURSOR:
			// Turn off the cursor when we're in fullscreen mode.
			if (g_active && g_ready && g_fullscreen) {
				SetCursor(NULL);
				return TRUE;
			}
			break;
        case WM_SIZE:
            // Check to see if we are losing our window...
            if (SIZE_MAXHIDE==wParam || SIZE_MINIMIZED==wParam)
                g_active = FALSE;
            else
                g_active = TRUE;
            break;
		default:
			if ((g_IPCMessage) &&
				(message == g_IPCMessage) &&
				(!(wParam & 0x80000000))) {
				if (ProcessIPCMessage(wParam & 0x7fff,
									  lParam))
					return TRUE;
			}
			break;
	}
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/////////////////////////////////////////////////////////////////////////////
//  This function permits apps to send messages to other apps.
void SendIPCMessage(u16 w, u32 d) {
	PostMessage(HWND_BROADCAST, g_IPCMessage, w | 0x80000000, d);
}

/////////////////////////////////////////////////////////////////////////////
//  Initialize the system.
static void __cdecl InitEngine(int nCmdShow) {
	HWND hWnd;
	WNDCLASS wc;
	int style;

	// Set up and register window class
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = g_hInst;
	wc.hIcon = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH )GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName = g_appName;
	wc.lpszClassName = g_appName;
	RegisterClass(&wc);

	#if 1
		// Create a window
		#if defined(DUMMY_BUFFER)
			style = WS_EX_OVERLAPPEDWINDOW;
		#else
			style = WS_EX_TOPMOST;
		#endif
		hWnd = CreateWindowEx(style,
							  g_appName,
							  g_appTitle,
							  WS_POPUP,
							  0,
							  0,
							  GetSystemMetrics(SM_CXSCREEN),
							  GetSystemMetrics(SM_CYSCREEN),
							  NULL,
							  NULL,
							  g_hInst,
							  NULL);
	#else
		hWnd = CreateWindowEx(0,
							  g_appName,
							  g_appTitle,
							  WS_OVERLAPPEDWINDOW,
							  CW_USEDEFAULT,
							  CW_USEDEFAULT,
							  1024+GetSystemMetrics(SM_CXSIZEFRAME)*2,
							  768+GetSystemMetrics(SM_CYSIZEFRAME)*2+GetSystemMetrics(SM_CYMENU),
							  NULL,
							  NULL,
							  g_hInst,
							  NULL);
	#endif
	if (!hWnd)
		exit(EXIT_FAILURE);
	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);
    // Save the window size/pos for switching modes
    GetWindowRect(hWnd, &g_rcWindow);
	SetFocus(hWnd);
	g_hWnd = hWnd;
	Initializer::InitializeAll();
}

/////////////////////////////////////////////////////////////////////////////
//  This thread is needed for windows apps:  the message pump.
static u8 WinMessagePump(void * v) {
	MSG msg;

	v = v;
	while (1) {
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) {
				ThreadKillAll();
				return 0;
			}
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		ThreadYield();
	}
	return 0;
}

static List < FString > s_engineArgs;
List < FString > g_args;

/////////////////////////////////////////////////////////////////////////////
//  Our (Win)Main function.
int APIENTRY WinMain(HINSTANCE hInst,
					 HINSTANCE hPrevInst,
					 LPSTR pCmdLine,
					 int nCmdShow) {
	ThreadStatus status;
	MType m, n;
	LPTSTR l = GetCommandLine();
	Bool q = FALSE;

	g_hInst = hInst;
	// Create the AGD message, if it doesn't already exist.
	g_IPCMessage = RegisterWindowMessage("AGD_IPC_MESSAGE");

	for (m=0; l[m]; ++m) {
		if (l[m] == '\"') {
			q = !q;
		}
		if (!q) {
			if (l[m] == ' ') {
				break;
			}
		}
		g_appPath[m] = l[m];
	}
	n=m;
	g_appPath[m] = '\0';
	if (g_appPath[0] == '\"') {
		for (m=1; g_appPath[m]; ++m) {
			g_appPath[m - 1] = g_appPath[m];
			if (g_appPath[m] == '\"') {
				g_appPath[m - 1] = '\0';
				break;
			}
		}
	}

	// Convert slashes.
	for (m=0; g_appPath[m]; ++m) {
		if (g_appPath[m] == '\\')
			g_appPath[m] = '/';
	}

	// Get rid of the terminal slash.
	--m;
	while (m && (g_appPath[m] != '/'))
		--m;
	g_appPath[m] = '\0';

	// That's our app path!  Now, let's process the remainder of the command
	// line.
	for (; l[n]; ++n) {
		if (l[n] == '-') {
			++n;
			if ((l[n] == 'e') || (l[n] == 'E')) {
				FString f;

				++n;
				q = FALSE;
				while ((l[n] != ' ') && (l[n] != '\t') && (l[n])) {
					if (l[n] == '\"') {
						q = !q;
					} else {
						f += l[n];
					}
					++n;
				}
				s_engineArgs += f;
			} else {
				FString f;

				q = FALSE;
				while ((l[n] != ' ') && (l[n] != '\t') && (l[n])) {
					if (l[n] == '\"') {
						q = !q;
					} else {
						f += l[n];
					}
					++n;
				}
				g_args += f;
			}
		}
	}

	#if defined(NDEBUG)
//		SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
	#endif

	InitEngine(nCmdShow);
	InitializeApp();

	// Let's launch all of our worker threads.
	ThreadLaunchThread(MainThread,
					   NULL,
					   g_mainThread,
					   status,
					   THREAD_FRAME_CYCLE,
					   0);
	ThreadLaunchThread(PtrGarbageCollectionThread,
					   NULL,
					   g_garbageThread,
					   status,
					   THREAD_FRAME_CYCLE,
					   0);
	ThreadLaunchThread(WinMessagePump,
					   NULL,
					   g_messagePumpThread,
					   status,
					   THREAD_FRAME_CYCLE,
					   0);
	ThreadEnterMultiThreadMode();

	ShutdownApp();
	DeInitializer::DeInitializeAll();
	#if defined(TRACK_MEMORY)
		MemoryDump();
	#endif

	g_soundManager.Shutdown();
	WarnBufferFlush();
	if (g_dbgFile) {
		fclose(g_dbgFile);
	}

	return 0;
}
