/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Okay, this sets up the graphics engine, sound engine, thread engine, etc.
and calls the app's entry point.

****************************************************************************/

#include "all.h"

//  These must be defined by the application.
extern u8 MainThread(void *);
extern u8 MousePoll(void *);
extern void InitializeApp(void);
extern void ShutdownApp(void);
extern char * g_appName;
extern char * g_appTitle;

//color g_redShift, g_greenShift, g_blueShift;
//u8 g_redMask, g_greenMask, g_blueMask;
u16 g_mouseX, g_mouseY;
u16 g_userWidth, g_userHeight;
u8 g_userDepth;
ThreadId g_mainThread, g_garbageThread, g_mouseThread;

extern FILE * g_dbgFile;

View g_monitor;

/////////////////////////////////////////////////////////////////////////////
//  Initialize the system.
static void InitEngine(void) {
	// Values normally derived from the registry.
	g_userWidth = 640;
	g_userHeight = 480;
	g_userDepth = 2;

	View::SetMode(g_userWidth, g_userHeight, g_userDepth);

	Initializer::InitializeAll();
}

/////////////////////////////////////////////////////////////////////////////
//  Our Main function.
int main(int argc, char ** argv) {
	ThreadStatus status;
	char * l = argv[0];
	Bool q = FALSE;

	for (m=0; l[m]; ++m) {
		if (l[m] == '\"') {
			q = !q;
		}
		if (!q) {
			if (l[m] == ' ') {
				break;
			}
		}
		g_appPath[m] = l[m];
	}
	g_appPath[m] = '\0';
	if (g_appPath[0] == '\"') {
		for (m=1; g_appPath[m]; ++m) {
			g_appPath[m - 1] = g_appPath[m];
			if (g_appPath[m] == '\"') {
				g_appPath[m - 1] = '\0';
				break;
			}
		}
	}

	// Convert slashes.
	for (m=0; g_appPath[m]; ++m) {
		if (g_appPath[m] == '\\')
			g_appPath[m] = '/';
	}

	// Kill a terminating slash.
	if (g_appPath[m - 1] == '/')
		g_appPath[m - 1] = '\0';

	InitEngine();
	InitializeApp();

	// Let's launch all of our worker threads.
	ThreadLaunchThread(MainThread,
					   NULL,
					   g_mainThread,
					   status,
					   THREAD_FRAME_CYCLE,
					   0);
	ThreadLaunchThread(PtrGarbageCollectionThread,
					   NULL,
					   g_garbageThread,
					   status,
					   THREAD_FRAME_CYCLE,
					   0);
	ThreadLaunchThread(MousePoll,
					   NULL,
					   g_mouseThread,
					   status,
					   THREAD_FRAME_CYCLE,
					   0);
	ThreadEnterMultiThreadMode();

	ShutdownApp();
	DeInitializer::DeInitializeAll();
	#if defined(TRACK_MEMORY)
		MemoryDump();
	#endif
	WarnBufferFlush();
	if (g_dbgFile) {
		fclose(g_dbgFile);
	}

	return 0;
}
