/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Maps.

****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
//  Construct a map.
template < class O >
inline
Map < O >::Map(void) {
	#if !defined(NDEBUG)
		m_isOpen = FALSE;
	#endif
	m_width = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Destroy a map.
template < class O >
inline
Map < O >::~Map(void) {
	if (m_width) {
		SafeDelete(m_elements[0]);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Clear out this map.
template < class O >
inline
void Map < O >::Destroy(void) {
	CHECK(!m_isOpen);
	if (m_width) {
		SafeDelete(m_elements[0]);
	}
	m_width = 0;
	m_elements.Destroy();
}

/////////////////////////////////////////////////////////////////////////////
//  Copy a map.
template < class O >
Map < O > & Map < O >::operator =(const Map < O > & m) {
	O * o;
	MType c, h;

	m_width = m.m_width;
	h = m.m_elements.Count();
	m_elements.Make(h, __FILE__, __LINE__);
	SafeNew(o, m_width * h, __FILE__, __LINE__);
	for (c=0; c<h; ++c) {
		MType d;

		m_elements[c] = o;
		o += m_width;
		for (d=0; d<m_width; ++d) {
			m_elements[c][d] = m.m_elements[c][d];
		}
	}
	#if !defined(NDEBUG)
		m_isOpen = m.m_isOpen;
	#endif
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Allocate a map of the indicated size.
template < class O >
void Map < O >::Allocate(MType width, MType height, const char * f, int l) {
	O * o;
	MType m;

	CHECK(!m_isOpen);
	if ((width != m_width) && (height != m_elements.Count())) {
		Destroy();
		SafeNew(o, width * height, f, l);
		m_elements.Make(height, f, l);
		for (m=0; m<height; ++m) {
			m_elements[m] = o;
			o += width;
		}
		m_width = width;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference a row in the map.
template < class O >
inline
MArray < O >
Map < O >::operator [](MType m) {
	CHECK(m_isOpen);
	CHECK(m < GetHeight());
	MArray < O > retval(m_width, m_elements[m]);
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference a row in the map, without modifying it.
template < class O >
inline
const MArray < O >
Map < O >::operator [](MType m) const {
	CHECK(m_isOpen);
	CHECK(m < GetHeight());
	MArray < O > retval(m_width, m_elements[m]);
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  How wide is the map.
template < class O >
inline
MType Map < O >::GetWidth(void) const {
	return m_width;
}

/////////////////////////////////////////////////////////////////////////////
//  How high is the map.
template < class O >
inline
MType Map < O >::GetHeight(void) const {
	return m_elements.Count();
}

#if !defined(NDEBUG)
	/////////////////////////////////////////////////////////////////////////
	//  When we're not in release mode, we make sure we're good citizens about
	// invoking Open and Close.
	template < class O >
	void Map < O >::Open(u8 option) {
		m_isOpen = TRUE;
	}

	/////////////////////////////////////////////////////////////////////////
	//  We're done writing.
	template < class O >
	void Map < O >::Close(void) {
		m_isOpen = FALSE;
	}
#else
	/////////////////////////////////////////////////////////////////////////
	//  In release mode, we assume the user has been a good citizen.
	template < class O >
	inline void Map < O >::Open(u8 option) {
	}

	/////////////////////////////////////////////////////////////////////////
	//  Done writing.
	template < class O >
	inline void Map < O >::Close(void) {
	}
#endif

