/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	This has all of our mouse support.  Really, we just need to track the
mouse and clicks.  We should make an OVERLAY primitive that we can use to
implement a mouse cursor.

****************************************************************************/

#ifndef MOUSE_H
	#define MOUSE_H

	/////////////////////////////////////////////////////////////////////////
	//	A region on the screen that can be clicked.
	class MouseHotspot {
		public:
			virtual ~MouseHotspot(void);

			u16 GetX(void) const;
			u16 GetY(void) const;
			u16 GetWidth(void) const;
			u16 GetHeight(void) const;

			void SetDims(u16 x, u16 y, u16 w, u16 h);

			virtual void Click(u16 x, u16 y);
			virtual void RClick(u16 x, u16 y);
		protected:
			Bool InRegion(u16 x, u16 y);
		private:
			u16 m_x;
			u16 m_y;
			u16 m_width;
			u16 m_height;
	};

	/////////////////////////////////////////////////////////////////////////
	//	A collection of mouse hotspots that can be activated or deactivated.
	class MouseHotspotGroup : public List < MouseHotspot * > {
		public:
			friend MESSAGE_PUMP;

			static void AddGroup(MouseHotspotGroup * p);
			static void RemoveGroup(MouseHotspotGroup * p);
		protected:
		private:
			static void Click(u16 x, u16 y);
			static void RClick(u16 x, u16 y);

			static List < MouseHotspotGroup * > s_allGroups;
	};

	u16 MouseGetX(void);
	u16 MouseGetY(void);
	void MouseDisplay(Bool offOn);
	Bool MouseIsVisible(void);

	#include "mouse.ipp"

#endif
