/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	2D primitives done here.

****************************************************************************/

#ifndef PRIM2D_H
	#define PRIM2D_H

	class View;
	class Bitmap;

	// I have to make this a templated class so I can accomodate rendering
	// options for different targets.
	struct RenderOptions {
		RenderOptions(void) : options(0), key(0), colorize(0), bg(0), blend(NULL) {};

		u8 options;
		color key;
		color colorize;
		color bg;
		Ptr < Bitmap > blend;
	};

	/////////////////////////////////////////////////////////////////////////
	//	Draw a line from one pair of screen coordinates to the other.
	extern void DrawLine(View & v,
						 color c,
						 u16 x1,
						 u16 y1,
						 u16 x2,
						 u16 y2);

	/////////////////////////////////////////////////////////////////////////
	//	Okay, here's the new class hierarchy:
	//
	//			 Target
	//				|
	//				|
	//				|
	//			Receptor
	//			/	   \
	//		   /		\
	//		View		RenderTexture
	//
	//	Targets are qualified recipients of any kind of rendering that is
	// done with the template parameter.  Receptors are Color Targets.
	template < class C >
	class Target : public Map < C > {
		friend class Overlay;
		public:
			u8 GetCurrentBuffer(void) const;

			virtual void Clear(C u, Rect * p = NULL)=0;
			virtual void Refresh(void);			// Flip pages.
			virtual C MapColor(real, real, real) const = 0;
		protected:
			void QueueOverlay(Overlay * o,
							  u16 x,
							  u16 y,
							  RenderOptions & ro);

			//	Render transparencies, overlays, etc.
			void Flush(void);
			void UndoOverlayDraw(void);

			// HEREHEREHERE we need the logic that says "this receptor blanks
			// everything out."
			class ORec {
				public:
					Overlay * overlay;
					U16 x;
					U16 y;
					RenderOptions ro;
			};

			Vec < List < ORec >, BUFFER_COUNT > m_overlays;
			u8 m_currentBuffer;
		private:
	};

	typedef Target < Color > Receptor;

	/////////////////////////////////////////////////////////////////////////
	//	This is an attribute class for all objects that are strictly 2D.
	template < class R, class S >
	class Object2D : public R {
		public:
			void SetPixelPosition(u16 x, u16 y);

			// Get my X and Y.
			u16 GetX(void);
			u16 GetY(void);

			virtual void Render(Receptor & m);
			S & GetOptions(void);
			void SetOptions(const S &);
		protected:
		private:
			S m_options;
			u16 m_x, m_y;
	};

	#include "prim2d.ipp"

#endif
