/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Inlines/templates for the various 2D primitive attribute classes.

****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
//  2D objects have a pixel position.  I may be changing the way this works
// when I get into coordinate spaces.
template < class R, class S >
inline void Object2D < R, S >::SetPixelPosition(u16 x, u16 y) {
	m_x = x;
	m_y = y;
}

/////////////////////////////////////////////////////////////////////////////
//  Get my X position.
template < class R, class S >
inline u16 Object2D < R, S >::GetX(void) {
	return m_x;
}

/////////////////////////////////////////////////////////////////////////////
//  Get my Y position.
template < class R, class S >
inline u16 Object2D < R, S >::GetY(void) {
	return m_y;
}

/////////////////////////////////////////////////////////////////////////////
//  Query the rendering options for this guy.  Ideally, this will enable a
// variety of different renderable objects.
template < class R, class S >
inline
S & Object2D < R, S >::GetOptions(void) {
	return m_options;
}

/////////////////////////////////////////////////////////////////////////////
//  Specify the rendering options for this guy.
template < class R, class S >
inline
void Object2D < R, S >::SetOptions(const S & s) {
	m_options = s;
}

/////////////////////////////////////////////////////////////////////////////
//  Render me on the indicated target.  HEREHEREHERE if we have trouble with
// this function, don't make it a member.
template < class R, class S >
void Object2D < R, S >::Render(Receptor & r) {
	R::Render(r, GetX(), GetY(), GetOptions());
}

/////////////////////////////////////////////////////////////////////////////
//  In a double-, triple-, etc. setting we need to know which buffer we're
// presently rendering to.
template < class C >
inline u8 Target < C >::GetCurrentBuffer(void) const {
	return m_currentBuffer;
}

/////////////////////////////////////////////////////////////////////////////
//  Queue an overlay for drawing at flush time.
template < class C >
void Target < C >::QueueOverlay(Overlay * o,
								u16 x,
								u16 y,
								RenderOptions & ro) {
	ORec & r = m_overlays[GetCurrentBuffer()].Grow();
	r.x = x;
	r.y = y;
	r.overlay = o;
	r.ro = ro;
}

/////////////////////////////////////////////////////////////////////////////
//  Flush the overlay buffer (render them).
template < class C >
void Target < C >::Flush(void) {
	MType m;
	List < ORec > & l = m_overlays[GetCurrentBuffer()];

	for (m=0; m<l.Count(); ++m) {
		l[m].overlay->ActuallyDraw(*this,
								   l[m].x,
								   l[m].y,
								   l[m].ro);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Undo the overlay drawing.  This has to be done in reverse order.
template < class C >
void Target < C >::UndoOverlayDraw(void) {
	MType m;
	List < ORec > & l = m_overlays[GetCurrentBuffer()];

	m = l.Count();
	while (m--) {
		l[m].overlay->Erase(*this, l[m].x, l[m].y);
	}
	l.Destroy();
}

/////////////////////////////////////////////////////////////////////////////
//  Effect all changes that have been made thus far (flip pages).
template < class C >
void Target < C >::Refresh(void) {
	Flush();
	++m_currentBuffer;
	if (m_currentBuffer == BUFFER_COUNT) {
		m_currentBuffer = 0;
	}
	UndoOverlayDraw();
}
