/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Some aspects of the pointer management are handled by these functions
here.

****************************************************************************/

#include "all.h"

#define MALLOC_GRAN							4096

struct ObjectRecord {
	void * object;
	void (*DestroyFunc)(void *);
};

static MType s_numberOfPointers, s_numberOfObjects;
static void *** s_pointers;
static ObjectRecord * s_objects;

/////////////////////////////////////////////////////////////////////////////
//  Add a pointer to the global list of all pointers.
MType PtrAddPointer(void * p) {
	MType retval;

	if (!(s_numberOfPointers & (MALLOC_GRAN - 1))) {
		s_pointers = (void ***)realloc(s_pointers,
									   (s_numberOfPointers + MALLOC_GRAN) *
									   sizeof(void **));
	}
	retval = s_numberOfPointers;
	s_pointers[s_numberOfPointers++] = (void **)p;
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Remove a pointer from the list of all pointers.  We put the last pointer
// in the list in the current position, and 
void PtrRemovePointer(MType which) {
	s_pointers[which] = s_pointers[--s_numberOfPointers];
	*((MType *)(s_pointers[which] + 1)) = which;
}

/////////////////////////////////////////////////////////////////////////////
//  Add an object to the global list.
void PtrAddObject(void * o, void (*DestroyFunc)(void *)) {
	CHECK(o);
	if (!(s_numberOfObjects & (MALLOC_GRAN - 1))) {
		s_objects = (ObjectRecord *)realloc(s_objects,
											(s_numberOfObjects + MALLOC_GRAN) *
											sizeof(ObjectRecord));
	}
	s_objects[s_numberOfObjects].object = o;
	s_objects[s_numberOfObjects++].DestroyFunc = DestroyFunc;
}

/////////////////////////////////////////////////////////////////////////////
//  This is a thread that looks for dead objects and destroys them.
u8 PtrGarbageCollectionThread(void * arg) {
	MType index=0;			// Acts like a static.

	while (1) {
		MType m, ub=s_numberOfObjects >> 5;

		if (!ub) {
			ub = s_numberOfObjects;
		}

		for (m=0; m<ub; ++m) {
			void * o;
			MType n;
	
			if (index >= s_numberOfObjects) {
				index = 0;
			}
			o = s_objects[index].object;
			for (n=0; n<s_numberOfPointers; ++n) {
				if (*(s_pointers[n]) == o) {
					break;
				}
			}
			if (n == s_numberOfPointers) {
				s_objects[index].DestroyFunc(o);
				free(o);
				s_objects[index] = s_objects[--s_numberOfObjects];
				ub--;
			} else {
				++index;
			}
		}
		ThreadYield();
	}
	return 0;
}

#if defined(TRACK_MEMORY)
	AllocRecord * g_allocRecords;
	MType g_maxMem;
	MType g_outstanding;

	/////////////////////////////////////////////////////////////////////////
	//  Do a dump of outstanding memory.  This will generally be called at
	// the end of program execution.
	void MemoryDump(void) {
		AllocRecord * a, * n;
		MType total = 0;

		Warn("===Outstanding memory allocations:\n");
		for (a=g_allocRecords; a; a=n) {
			n = a->next;
			total += a->size;
			Warn("------%x bytes from %s, line %d.\n",
				 a->size,
				 a->f,
				 a->l);
		}
		Warn("------ Total outstanding memory:  %x.\n", total);
		Warn("------ Max allocated at an instant:  %x.\n", g_maxMem);
		Warn("===End of memory dump.\n");
	}

#endif

