/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	This handles all of our sound.

****************************************************************************/

#include "all.h"

LPDIRECTSOUND       g_pDS            = NULL;
LPDIRECTSOUNDNOTIFY g_pDSNotify      = NULL;

static MType s_bufSize;
static LPDIRECTSOUNDBUFFER s_primary, s_secondary;

static TString s_soundFailure("Unable to initialize DirectSound.");

SoundManager g_soundManager;

/////////////////////////////////////////////////////////////////////////////
//  Insert the indicated array of sounds into the sound manager.
void SoundManager::Insert(u8 voice, MType count, Sound * s) {
	LPVOID ptr1, ptr2;
	DWORD size1, size2;
	HRESULT hr;
	int ret=0;

	hr=s_secondary->Lock(voice * s_bufSize * 4,
						 s_bufSize * 4,
						 &ptr1,
						 &size1,
						 &ptr2,
						 &size2,
						 0);
	if (hr != DS_OK) {
		return;
	}

	if (ptr1!=NULL)	{
		Sound * convsnd16 = (Sound *)ptr1;
		MType j;

		if (s) {
			for (j=0; j<s_bufSize * 2; ++j) {
				convsnd16[j] = s[j];
			}
		}
		s_secondary->Unlock(ptr1, size1, ptr2, size2);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Initialize the sound manager.
void SoundManager::Initialize(void) {
	HRESULT hr;
    DSBUFFERDESC dsbd;
    PCMWAVEFORMAT pcmwf;
    WAVEFORMATEX wf;

	s_bufSize = SOUND_RATE / 60;
	if (s_bufSize < 16) {
		s_bufSize = 16;
	}
	if (FAILED(hr=DirectSoundCreate(NULL, &g_pDS, NULL))) {
		::Shutdown(s_soundFailure);
	}
	if (FAILED(hr=g_pDS->SetCooperativeLevel(g_hWnd, DSSCL_EXCLUSIVE))) {
		::Shutdown(s_soundFailure);
	}

	memset(&pcmwf, 0, sizeof(pcmwf));
	memset(&wf, 0, sizeof(wf));
	pcmwf.wBitsPerSample = 16;
	pcmwf.wf.wFormatTag = WAVE_FORMAT_PCM; 
	pcmwf.wf.nChannels = 2; 
	pcmwf.wf.nSamplesPerSec = SOUND_RATE;
	pcmwf.wf.nBlockAlign = pcmwf.wf.nChannels * (pcmwf.wBitsPerSample / 8); 
	pcmwf.wf.nAvgBytesPerSec = pcmwf.wf.nSamplesPerSec *
							   pcmwf.wf.nBlockAlign; 

	wf.wFormatTag = WAVE_FORMAT_PCM;
	wf.nChannels = pcmwf.wf.nChannels;
	wf.nSamplesPerSec = pcmwf.wf.nSamplesPerSec;
	wf.wBitsPerSample = pcmwf.wBitsPerSample;
	wf.nBlockAlign = pcmwf.wf.nBlockAlign;
	wf.nAvgBytesPerSec = pcmwf.wf.nAvgBytesPerSec;
	wf.cbSize = sizeof(WAVEFORMATEX);


	// Get the primary buffer.
	memset(&dsbd, 0, sizeof(dsbd));
	dsbd.dwSize        = sizeof(DSBUFFERDESC);
	dsbd.dwFlags       = DSBCAPS_PRIMARYBUFFER;

	if (FAILED(hr = g_pDS->CreateSoundBuffer(&dsbd, &s_primary, NULL ))) {
	    ::Shutdown(s_soundFailure);
	}
	
	// Set primary buffer format to 22kHz and 16-bit output.
	if (FAILED(hr = s_primary->SetFormat(&wf))) {
	    ::Shutdown(s_soundFailure);
	}
	if (FAILED(hr = s_primary->Play(0, 0, DSBPLAY_LOOPING))) {
		::Shutdown(s_soundFailure);
	}

	// Now, let's make our secondary buffer.
	memset(&dsbd, 0, sizeof(DSBUFFERDESC));
	dsbd.dwSize = sizeof(DSBUFFERDESC);
	dsbd.dwFlags = DSBCAPS_CTRLPAN |
				   DSBCAPS_CTRLVOLUME |
				   DSBCAPS_CTRLFREQUENCY |
				   DSBCAPS_GETCURRENTPOSITION2 |
				   DSBCAPS_STICKYFOCUS;
	dsbd.dwBufferBytes = s_bufSize * SOUND_NUMBER_OF_VOICES * 4; 
	dsbd.lpwfxFormat = (LPWAVEFORMATEX)&pcmwf;

	if (FAILED(hr = g_pDS->CreateSoundBuffer(&dsbd, &s_secondary, NULL))) {
		::Shutdown(s_soundFailure);
	}
	if (FAILED(hr = s_secondary->Play(0, 0, DSBPLAY_LOOPING))) {
		::Shutdown(s_soundFailure);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  We're all done with the sound.
void SoundManager::Shutdown(void) {
    // Release DirectSound interfaces.
    if (s_primary) {
    	s_primary->Release();
    	s_primary = NULL;
    }
    if (s_secondary) {
    	s_secondary->Release();
    	s_secondary = NULL;
    }
    if (g_pDS) {
    	g_pDS->Release();
    	g_pDS = NULL;
    }
}
