/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of all of our text functions.

****************************************************************************/

#include "all.h"

/////////////////////////////////////////////////////////////////////////////
//  This holds all the information we need about a font.
struct Font {
	u16 width;
	u16 height;
	Array < Bitmap > images;
};

static List < Font > s_allFonts;
static U16 s_tabs[TEXT_NUM_TABS];

/////////////////////////////////////////////////////////////////////////////
//  Load a text font.  Presently, we have international support, but we don't
// yet have proportional font support or blending support.
FontHandle TextLoadFixedFont(const FString & name,
							 u16 fw,
							 u16 fh,
							 u16 numberOfCharacters) {
	FontHandle retval;
	u16 i, j, x;
	Map < Color > m;
	List < Bitmap > images;

	g_monitor.ImportFile(name, m);

	// Okay, that retrieved our data.  Now, we want to carve out the
	// individual font images.
	Font & f = s_allFonts.Grow();
	retval = s_allFonts.Count();

	f.width = fw;
	f.height = fh;

	for (j=0, x=0; j<m.GetHeight(); j+=fh) {
		for (i=0; i<m.GetWidth(); i+=fw, ++x) {
			images.Grow().ExtractFrom(m, i, j, fw, fh);
		}
	}
	f.images = images;

	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  This will actually write text to the view.
void TextWrite(Receptor & v,
			   FontHandle f,
			   u16 x,
			   u16 y,
			   color d,
			   const FString & i,
			   u8 options) {
	Font & font = s_allFonts[--f];
	MType m, n;
	color c = d;
	color bg=0;

	for (m=0; m<i.Length(); ++m) {
		if (i[m] < TEXT_NUM_CONTROL_CODES) {
			switch(i[m]) {
				case TEXT_COLOR_BLACK:
					c = v.MapColor(0, 0, 0);
					break;
				case TEXT_COLOR_RED:
					c = v.MapColor(1, 0, 0);
					break;
				case TEXT_COLOR_GREEN:
					c = v.MapColor(0, 1, 0);
					break;
				case TEXT_COLOR_BLUE:
					c = v.MapColor(0, 0, 1);
					break;
				case TEXT_COLOR_ORANGE:
					c = v.MapColor(1, 0.5, 0);
					break;
				case TEXT_COLOR_YELLOW:
					c = v.MapColor(1, 1, 0);
					break;
				case TEXT_COLOR_PURPLE:
					c = v.MapColor(1, 0, 1);
					break;
				case TEXT_COLOR_WHITE:
					c = v.MapColor(1, 1, 1);
					break;
				case TEXT_COLOR_DEFAULT:
					c = d;
					break;
				case TEXT_COLOR_SPECIFIED:
					++m;
					c = v.MapColor((i[m] & 0x1f) / 31.0,
								   ((i[m] >> 5) & 0x1f) / 31.0,
								   ((i[m] >> 10) & 0x1f) / 31.0);
					break;
				case TEXT_BGCOLOR_SPECIFIED:
					++m;
					bg = v.MapColor((i[m] & 0x1f) / 31.0,
								    ((i[m] >> 5) & 0x1f) / 31.0,
								    ((i[m] >> 10) & 0x1f) / 31.0);
					break;
				case TEXT_TAB:
					for (n=0; n<TEXT_NUM_TABS; ++n) {
						if (s_tabs[n] > x) {
							x = s_tabs[n];
							break;
						}
					}
					break;
			}
		} else {
			RenderOptions r;

			r.options = options;
			r.bg = bg;
			r.colorize = c;
			r.key = 0;
			if (i[m] < font.images.Count()) {
				font.images[i[m]].Render(v, x, y, r);
			} else {
				font.images['?'].Render(v, x, y, r);
			}
			x += font.width;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  How wide is the widest character of the font.
u16 TextGetFontWidth(FontHandle f) {
	return s_allFonts[f - 1].width;
}
/////////////////////////////////////////////////////////////////////////////
//  How high is this font.
u16 TextGetFontHeight(FontHandle f) {
	return s_allFonts[f - 1].height;
}

/////////////////////////////////////////////////////////////////////////////
//  Set the indicated tab stop.
void TextSetTab(u8 tab, u16 indent) {
	CHECK(tab < TEXT_NUM_TABS);
	s_tabs[tab] = indent;
}

/////////////////////////////////////////////////////////////////////////////
//  Clear all the font stuff from memory.
static void TextShutdown(void) {
	s_allFonts.Destroy();
}

static DeInitializer s_textShutdown(TextShutdown);
