;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;	Written by John Marco Panettiere for John Marco Panettiere
;	Copyright 1998, 1999, 2000 by John Marco Panettiere.
;	All rights reserved.
;
;	I need to be able to have precise control over the stack frame to be
; able to make independent tasks.  For instance, I cannot have any local
; variables on the stack!  Therefore, I will code this in assembly.
;	Okay, setjmp and longjmp seem to have these really weird OS implications,
; so I will code up my own routines.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

[SECTION .text USE32]

%define MAX_THREADS							18
%define STACK_SIZE							16 * 1024

[GLOBAL _ThreadPrepareThreadLaunches]
[GLOBAL _ThreadSaveStackState]
[GLOBAL _ThreadRestoreStackState]
[GLOBAL _ThreadSetJmp]
[GLOBAL _ThreadLongJmp]
[GLOBAL _GetCPUSeconds]
[GLOBAL _GetCPUTicks]

[EXTERN _ThreadKillThread]
%ifdef DOS
[EXTERN g_processorSpeedInverse]
%define _g_processorSpeedInverse g_processorSpeedInverse
%else
[EXTERN _g_processorSpeedInverse]
%endif

_GetCPUSeconds:
	rdtsc
	mov	[s_scratch + 4],edx
	mov	[s_scratch],eax
	fild qword [s_scratch]
	fmul dword [_g_processorSpeedInverse]
%ifdef DOS
	fstp qword [s_scratch]
	mov eax,s_scratch
%endif
	ret

_GetCPUTicks:
	rdtsc
	mov [esi],eax
	mov [esi + 4],edx
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  When we suspend a thread, we preserve its entire stack state.  Format:
; void ThreadSaveStackState(Thread &, U8 * buffer);
_ThreadSaveStackState:
	push esi
	push edi

	mov edx,[esp + 12]
	mov ecx,STACK_SIZE / 4

	mov esi,[esp + 16]
	mov edi,[edx + StackBase]

	rep movsd

	pop edi
	pop esi

	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  Restore the stack state of the thread.
_ThreadRestoreStackState:
	push esi
	push edi

	mov edx,[esp + 12]
	mov ecx,STACK_SIZE / 4

	mov esi,[edx + StackBase]
	mov edi,[edx + 16]

	rep movsd

	pop edi
	pop esi

	ret

				struc	Thread
LastTime:		resd	1
Ticks:			resd	1
NetTime:		resd	1
EntryPoint:		resd	1
EntryArg:		resd	1
LastPosition:	resd	22
ThreadStatus:	resd	1
StackBase:		resd	1
Priority:		resd	1
				endstruc

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  By only using registers to represent local variables, this should enable
; us to set up a proper task switching location.
_ThreadPrepareThreadLaunches:
	push ebx
	push ebp
	push edi
	push esi

	mov ebp,[esp + 20]						; Threads.
	mov esi,[esp + 24]						; s_id.

	sub ebx,ebx
	mov edi,esp

TPTL1:
	sub esp,STACK_SIZE						; Local stack.
	lea eax,[ebp + LastPosition]

	mov [ebp + StackBase],esp
	push eax
	call _ThreadSetJmp

	add esp,4
	test eax,eax
	jz TPTL2

	push dword [ebp + EntryArg]
	call dword [ebp + EntryPoint]

; The register dependencies below are based on registers that are preserved
; across function calls.
	mov eax,[ebx*4 + esi]
	mov [esp],eax

	call _ThreadKillThread

	mov [esp],ebx
;	call ThreadNextThread
	add esp,4

TPTL2:
	inc ebx
	add ebp,Thread_size

	cmp ebx,MAX_THREADS
	jnz TPTL1

	mov esp,edi

	pop esi
	pop edi
	pop ebp
	pop ebx

	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  Save this location as a future return point.  Return 0.
; What do we save when we call this function?  Let's save all of the
; registers, including the FP registers.
_ThreadSetJmp:
	mov eax,[esp + 4]
	mov ecx,[esp]
	mov [eax + 0x00],ecx
	mov [eax + 0x04],ebx
	mov [eax + 0x08],esi
	mov [eax + 0x0c],edi
	mov [eax + 0x10],ebp
	lea ecx,[esp + 4]
	mov [eax + 0x14],ecx
	xor eax,eax
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  Execute a long jump to the specified buffer.
_ThreadLongJmp:
	mov ecx,[esp + 4]
	mov eax,[esp + 8]
	mov ebx,[ecx + 0x04]
	mov esi,[ecx + 0x08]
	mov edi,[ecx + 0x0c]
	mov ebp,[ecx + 0x10]
	mov esp,[ecx + 0x14]
	jmp [ecx]

[SECTION .bss USE32]
s_scratch:
	resq	1
