/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	This will be my new thread library.  Unfortunately, it seems that not
enough standards are in place for preemptive multithreading.  I will use
this library to insulate the caller from the DOS/Unix/Win32 details, and
whether or not actual multiprocessing is employed.
	24Oct98, 04:06:  A lot of progress made in this file!  Need to figure
out how to automate prologue (or epilogue) code that yields.

****************************************************************************/

#ifndef THREADS_H
	#define THREADS_H

	// Status flags.
	enum {
		// These flags are set by the engine.
		THREAD_WAITING_TO_START				=	0x00000001,
		THREAD_RUNNING						=	0x00000002,
		THREAD_NOT_A_THREAD					=	0x00000004,
		THREAD_SLEEPING						=	0x00000008,
		THREAD_SUSPENDED					=	0x00000010,

		THREAD_TYPE_SHIFT					=	0x00000008,

		THREAD_FRAME_CYCLE					=	0x00000000,
		THREAD_CRITICAL,
		THREAD_TIMESLICED,
		THREAD_PRIORITIZED
	};

	// Return codes.
	enum {
		THREAD_OK							=	0,	// Call succeeded.
		THREAD_MEM,									// Insufficient memory.
		THREAD_INVALID,								// Invalid thread id.
	};

	enum {
		THREAD_MAX_THREADS					=	16,
		THREAD_STACK_SIZE					=	16 * 1024
	};


	class ThreadSaveBuffer;

	extern "C" {
		extern real g_processorSpeedInverse;
		extern u8 __cdecl ThreadKillThread(ThreadId);
	}

	template < class O >
	u8 ThreadLaunchThread(ThreadEntryPoint,
						  Ptr < O > &,
						  ThreadId &,
						  ThreadStatus &,
						  u8,
						  MType);
	u8 ThreadLaunchThread(ThreadEntryPoint,
						  void *,
						  ThreadId &,
						  ThreadStatus &,
						  u8,
						  MType);
	extern u8 ThreadGetCurrentThread(ThreadId &);
	extern u8 ThreadGetThreadStatus(ThreadId,
									ThreadStatus &);

	extern u8 ThreadMakeFrameCycled(ThreadId);
	extern u8 ThreadMakeCritical(ThreadId);
	extern u8 ThreadMakeTimesliced(ThreadId);
	extern u8 ThreadMakePrioritized(ThreadId);

	extern u8 ThreadAddTimeslice(ThreadId,
								 real timeslice);
	extern u8 ThreadBoostPriority(ThreadId);
	extern u8 ThreadLowerPriority(ThreadId);

	extern u8 ThreadKillZombie(ThreadId);
	extern u8 ThreadKillThreadAndZombie(ThreadId);
	// Removes from active queue.
	extern u8 ThreadSuspendThread(ThreadId, ThreadSaveBuffer &);
	extern u8 ThreadResumeThread(ThreadId &,
								 ThreadSaveBuffer &);
	// Does not remove it from active queue.
	extern u8 ThreadSuspend(ThreadId id, Bool suspend);
	extern u8 ThreadGetCPUTime(ThreadId, real &);

	#define ThreadGetStackSize(x) (THREAD_STACK_SIZE)

	extern void ThreadYield(void);
	extern void ThreadEnterMultiThreadMode(void);
	extern void ThreadKillAll(void);

	/////////////////////////////////////////////////////////////////
	//  For internal use.

	#define THREAD_MAX_PRIORITY				0x10
	#define THREAD_MIN_PRIORITY				0x01

	struct ThreadJmpBuf {
		U32 addr;					 // eip
		U32 regs[5];				 // ebx, esi, edi, ebp, esp.
		U32 fp[16];					 // fpu / mmx.
	};

	struct Thread {
		real lastTime;				 // When I started executing.
		real allotment;				 // Time I've been allotted.
		real netCPUTime;			 // How many ticks have passed.
		ThreadEntryPoint entryPoint; // Where I start running.
		void * entryArg;			 // Argument I get when I run.
		ThreadJmpBuf lastPosition;	 // Where I last left off.
		ThreadStatus threadStatus;	 // Attributes.
		U8 * stackBase;				 // Where my stack starts.
		u8 priority;				 // How important am I.
	};
	
	class ThreadSaveBuffer {
		public:
			ThreadSaveBuffer(ThreadId & t);
			~ThreadSaveBuffer(void) {
				if (m_stackSaveBuffer)
					SafeDelete(m_stackSaveBuffer);
			}
			U8 * GetBuffer(void) {
				return m_stackSaveBuffer;
			}
			void CopyThreadInfo(Thread & thread) {
				memcpy(&m_thread, &thread, sizeof(thread));
			}
			void RestoreThreadInfo(Thread & thread) {
				U8 * t = thread.stackBase;

				memcpy(&thread, &m_thread, sizeof(thread));
				thread.stackBase = t;
			}
			real GetCPUTime(void) {
				return (real)m_thread.netCPUTime;
			}
		protected:
		private:
			ThreadSaveBuffer(void);

			ThreadStatus m_status;
			Thread m_thread;
			U8 * m_stackSaveBuffer;
	};

	// Thread mutex constants.
	enum {
		THREAD_MUTEX_ASYNC=0,				// If already locked, note this.
		THREAD_MUTEX_WAIT,					// Wait 'til locked.

		THREAD_MUTEX_SUCCESSFUL,			// Lock succeeded.
		THREAD_MUTEX_ALREADY_LOCKED,		// It's locked.
	};

	/////////////////////////////////////////////////////////////////
	//	Mutexes serialize access to an object.  If thread A locks a
	// mutex, thread B is shut out from it.  A can call lock as many
	// times as it wants, then unlock it.
	template < class Item > class ThreadMutex {
		public:
			ThreadMutex(void) {
				m_status = 0;
			}
			u8 Lock(u8 mutexOption) {
				u8 retval;
				ThreadId current;

				ThreadGetCurrentThread(current);

				switch(mutexOption) {
					case THREAD_MUTEX_ASYNC:
						if (m_status &&
							(m_status != current)) {
							return THREAD_MUTEX_ALREADY_LOCKED;
						}
						break;
					case THREAD_MUTEX_WAIT:
						while (m_status &&
							   (m_status != current)) {
							ThreadYield();
						}
						break;
				}
				m_status = current;
				return THREAD_MUTEX_SUCCESSFUL;
			}
			/////////////////////////////////////////////////////////
			//  Can't unlock if I'm not currently locking it.
			void Unlock(void) {
				ThreadId current;

				ThreadGetCurrentThread(current);
				if (m_status == current) {
					m_status = 0;
				}
			}
			Item & operator ()(void) {
				return m_item;
			}
		protected:
		private:
			ThreadId m_status;
			Item m_item;
	};

/****************************************************************************
*****************************************************************************

			D	O	C	U	M	E	N	T	A	T	I	O	N

		A thread is an active process.  A ThreadSaveBuffer holds threads
	that are running but not presently active (don't necessarily get a
	timeslice this frame).  Threads recognize the existence of a frame that
	gets partitioned over several different kinds of threads that execute
	in that frame:

		THREAD_FRAME_CYCLE threads execute once per frame.
		THREAD_CRITICAL threads execute until completion.
		THREAD_TIMESLICED threads execute for a pre-specified timeslice.
		THREAD_PRIORITIZED get timeslices based on a priority system.

		During a given frame, the thread engine does the following:
		(1)  Executes all THREAD_FRAME_CYCLE threads, letting them run until
		they yield or terminate.
		(2)  Executes all THREAD_CRITICAL threads until they terminate.
		(3)  Figures out how much time is left in the frame to sustain a pre-
		specified framerate.  Call this value t.
		(4)  Picks a number that is some function of the total priority of
		all THREAD_PRIORITIZED threads.  When this value is small, we allot
		a lot of t to THREAD_TIMESLICED threads.  When the value is large,
		more time gets allotted to THREAD_PRIORITIZED threads.
		(5)  Take the allotment of the THREAD_PRIORITIZED threads and divide
		it among those threads using their priorities as a weight.
		(6)  Execute the THREAD_PRIORITIZED threads using their allotments.
		(7)  Sort the THREAD_TIMESLICED threads based on their allotments.
		(8)  Execute the THREAD_TIMESLICED threads until the duration of the
		frame matches the framerate.
		(9)  Start a new frame.

	Types:
		ThreadEntryPoint - This is where a thread begins its execution.
		ThreadId - This uniquely identifies a thread.
		ThreadPriority - Objects used to set the priority level of a thread.
			At some point, we have to worry about making threads critical,
			etc.
		ThreadStatus - Has fields indicating the current status of a thread.
		ThreadMutex - Serializes accesses to an object.

	Functions:
		ThreadOpen - Needed to initialize the thread manager.  Ignored if
			already open.
		ThreadEnterMultithreadMode - transfers control to the thread manager.
			Systems supporting preemptive multithreading ignore this call.
		ThreadLaunchThread - Used to launch a new thread.  The parameters are
			the function that the thread will start on (when this exits, the
			thread terminates), a reference accepting the thread id, the
			initial priority of the thread, flags describing the thread, and
			the stack size needed for the thread.
		ThreadGetThreadPriority - Query the current priority of the thread.
		ThreadSetThreadPriority - Set the priority of the thread.
		ThreadGetThreadStatus - Get the current status of the thread.
		ThreadSetThreadStatus - Set the current status of the thread.
		ThreadKillThread - Abort the current thread.
		ThreadKillZombie - If the zombie is tying up resources, kill it.
		ThreadKillThreadAndZombie - Kill them both.
		ThreadMutex::Lock() - Reserve access to an object.
		ThreadMutex::Unlock() - Relinquish control of the object.
	The polling macro is there to allow us to accomodate non-preemptive
multi-threading environments, like DPMI.

*****************************************************************************
****************************************************************************/

	#include "threads.ipp"

	extern ThreadId g_mainThread,
					g_garbageThread,
					g_messagePumpThread;

#endif
