/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere
	Copyright 1996, 1997 John Marco Panettiere

	This the DOS implementation of my VIEW class.

****************************************************************************/

#include "all.h"

extern HRESULT __cdecl InitFail(HWND hWnd,
								HRESULT hRet,
								LPCTSTR szError,
								...);
ActiveGraphic * ActiveGraphic::s_allActiveGraphics;

RECT                    g_rcWindow;             // Saves the window size & pos.
RECT                    g_rcViewport;           // Pos. & size to blt from
RECT                    g_rcScreen;             // Screen pos. for blt
Bool g_active = FALSE;
Bool g_ready = FALSE;
Bool g_fullscreen = TRUE;

/////////////////////////////////////////////////////////////////////////////
//  A "macro" to convert from color to d3d RGB.  HEREHEREHERE not needed yet!
inline DWORD ColorToDDrawRGB(color c) {
	unsigned r, g, b;

	r = (c >> g_redShift) * 255 / g_redMask;
	g = (c >> g_greenShift) * 255 / g_greenMask;
	b = (c >> g_blueShift) * 255 / g_blueMask;

	return 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Destroy the view.  Because the display memory is managed externally, we
// manually clobber the pointer so that the receptor destructor doesn't free
// it up.
View::~View(void) {
	if (m_elements.Count()) {
		m_elements[0] = NULL;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Create the view.
View::View(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  Initialize the main view.
void View::Initialize(void) {
	MType m;

	g_monitor.m_elements.Make(g_ddrawHeight, __FILE__, __LINE__);
	g_monitor.m_width = g_ddrawWidth;
	for (m=0; m<BUFFER_COUNT; ++m) {
		g_monitor.m_overlays[m].Destroy();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Flip the pages.
void View::Refresh(void) {
	Flush();
	#if !defined(DUMMY_BUFFER)
		HRESULT hRet;

		if (g_fullscreen) {
			if (g_active) {
				while (TRUE) {
					hRet = g_pDDSPrimary->Flip(NULL, 0);
					if (hRet == DD_OK)
						break;
					if (hRet == DDERR_SURFACELOST) {
						hRet = g_pDDSPrimary->Restore();
						if (hRet != DD_OK)
							break;
					}
					if (hRet != DDERR_WASSTILLDRAWING)
						break;
				}
			}
		} else {
			if (g_active) {
				while (TRUE) {
					hRet = g_pDDSPrimary->Blt(&g_rcScreen,
											  g_pDDSBack,
											  &g_rcViewport,
											  DDBLT_WAIT,
											  NULL);
					if (hRet == DD_OK )
						break;
					if (hRet == DDERR_SURFACELOST) {
						hRet = g_pDDSPrimary->Restore();
						if (hRet != DD_OK )
							break;
					}
					if (hRet != DDERR_WASSTILLDRAWING )
						break;
				}
			}
		}
	#endif
	++m_currentBuffer;
	if (m_currentBuffer == BUFFER_COUNT) {
		m_currentBuffer = 0;
	}
	UndoOverlayDraw();
}

/////////////////////////////////////////////////////////////////////////////
//  Clear the back buffer.  In the future, we check for hardware
// acceleration.  Also, must check for m_isOpen.
void View::Clear(Color c, Rect * r) {
	color u = c;

	#if !defined(DUMMY_BUFFER)
		DDBLTFX ddbltfx;
		RECT rect;

		// Use the blter to do a color fill to clear the back buffer
		ZeroMemory(&ddbltfx, sizeof(ddbltfx));
		ddbltfx.dwSize = sizeof(ddbltfx);
		ddbltfx.dwFillColor = u;
		if (r) {
			rect.top = r->y;
			rect.left = r->x;
			rect.right = r->x + r->w;
			rect.bottom = r->y + r->h;
			g_pDDSBack->Blt(&rect,
							NULL,
							NULL,
							DDBLT_COLORFILL | DDBLT_WAIT,
							&ddbltfx);
		} else {
			g_pDDSBack->Blt(NULL,
							NULL,
							NULL,
							DDBLT_COLORFILL | DDBLT_WAIT,
							&ddbltfx);
		}
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Convert from the color to the r, g, b values.
void View::MapColor(Color c, real & r, real & g, real & b) const {
	color d;

	d = (c >> g_redShift) & g_redMask;
	r = (real)d / (real)g_redMask;
	d = (c >> g_greenShift) & g_greenMask;
	g = (real)d / (real)g_greenMask;
	d = (c >> g_blueShift) & g_blueMask;
	b = (real)d / (real)g_blueMask;
}

/////////////////////////////////////////////////////////////////////////////
//  Convert the indicated r, g, b values to a platform-specific color.
Color View::MapColor(real r, real g, real b) const {
	color retval;

	retval  = (color)(r * g_redMask) << g_redShift;
	retval |= (color)(g * g_greenMask) << g_greenShift;
	retval |= (color)(b * g_blueMask) << g_blueShift;
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  This is the non-inlined portion of locking the back buffer.
void View::OLLock(void) {
	#if !defined(DUMMY_BUFFER)
		DDSURFACEDESC2 ddsd;
		u16 w;
		HRESULT r;
		Bool needRedraw = FALSE;
	
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		while (1) {
			r = g_pDDSBack->Lock(NULL, &ddsd, 0, NULL);
			if (r == DDERR_SURFACELOST) {
				ThreadYield();
				g_pDDSPrimary->Restore();
				needRedraw = TRUE;
				continue;
			} else if ((r == DDERR_SURFACEBUSY) ||
					   (r == DDERR_WASSTILLDRAWING)) {
				ThreadYield();
				continue;
			} else {
				break;
			}
		}
		if (needRedraw)
			ActiveGraphic::RedrawAll();
		for (w=0; w<g_ddrawHeight; ++w) {
			m_elements[w] = (Color *)((U8 *)ddsd.lpSurface + w * ddsd.lPitch);
		}
	#else
		static Color * s_buffer;
		static u16 s_width, s_height;

		if ((g_ddrawWidth != s_width) ||
			(g_ddrawHeight != s_height)) {
			MType m;

			SafeNew(s_buffer,
					g_ddrawWidth * g_ddrawHeight,
					__FILE__,
					__LINE__);
			for (m=0; m<g_ddrawHeight; ++m) {
				m_elements[m] = s_buffer + m * g_ddrawWidth;
			}
			s_width = g_ddrawWidth;
			s_height = g_ddrawHeight;
		}
	#endif
	WarnSetBuffering(TRUE);
	m_locked = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Non-inlined portion of the unlock function.
void View::OLUnlock(void) {
	#if !defined(DUMMY_BUFFER)
		g_pDDSBack->Unlock(NULL);
	#endif
	m_locked = FALSE;
	WarnSetBuffering(FALSE);
	WarnBufferFlush();
}

extern void ReleaseMostObjects(HWND hWnd);

/////////////////////////////////////////////////////////////////////////////
//  Taken from a DirectDraw example.
static HRESULT InitSurfaces(HWND hWnd,
							Bool windowed,
							u16 width,
							u16 height,
							u8 depth) {
    HRESULT		        hRet;
    DDSURFACEDESC2      ddsd;
    DDSCAPS2            ddscaps;
    LPDIRECTDRAWCLIPPER pClipper;

    if (windowed) {
        // Get normal windowed mode
        hRet = g_pDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "SetCooperativeLevel FAILED");

    	// Get the dimensions of the viewport and screen bounds
    	GetClientRect(hWnd, &g_rcViewport);
    	GetClientRect(hWnd, &g_rcScreen);
    	ClientToScreen(hWnd, (POINT*)&g_rcScreen.left);
    	ClientToScreen(hWnd, (POINT*)&g_rcScreen.right);

        // Create the primary surface
        ZeroMemory(&ddsd,sizeof(ddsd));
        ddsd.dwSize = sizeof(ddsd);
        ddsd.dwFlags = DDSD_CAPS;
        ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
        hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSPrimary, NULL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "CreateSurface FAILED");

        // Create a clipper object since this is for a Windowed render
        hRet = g_pDD->CreateClipper(0, &pClipper, NULL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "CreateClipper FAILED");

        // Associate the clipper with the window
        pClipper->SetHWnd(0, hWnd);
        g_pDDSPrimary->SetClipper(pClipper);
        pClipper->Release();
        pClipper = NULL;

        // Get the backbuffer. For fullscreen mode, the backbuffer was created
        // along with the primary, but windowed mode still needs to create one.
        ddsd.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
        ddsd.dwWidth        = width;
        ddsd.dwHeight       = height;
        ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
        hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSBack, NULL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "CreateSurface2 FAILED");
		// Get the ever-important pitch information.
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		hRet = g_pDDSPrimary->GetSurfaceDesc(&ddsd);
		if (hRet != DD_OK) {
			return FALSE;
		}
		g_ddrawWidth = ddsd.dwWidth;
		g_ddrawHeight = ddsd.dwHeight;
		g_ddrawPitch = ddsd.lPitch / sizeof(Color);
		g_ddrawDepth = depth;
    } else {
        // Get exclusive mode
        hRet = g_pDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE |
                                                DDSCL_FULLSCREEN);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "SetCooperativeLevel FAILED");

        hRet = g_pDD->SetDisplayMode(width,
        							 height,
        							 depth * 8,
        							 0,
        							 0);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "SetDisplayMode FAILED");

    	// Get the dimensions of the viewport and screen bounds
    	// Store the rectangle which contains the renderer
    	SetRect(&g_rcViewport, 0, 0, width, height);
    	memcpy(&g_rcScreen, &g_rcViewport, sizeof(RECT));

        // Create the primary surface with 1 back buffer
        ZeroMemory(&ddsd,sizeof(ddsd));
        ddsd.dwSize = sizeof(ddsd);
        ddsd.dwFlags = DDSD_CAPS |
                       DDSD_BACKBUFFERCOUNT;
        ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                			  DDSCAPS_FLIP |
                			  DDSCAPS_COMPLEX;
        ddsd.dwBackBufferCount = 1;
        hRet = g_pDD->CreateSurface( &ddsd, &g_pDDSPrimary, NULL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "CreateSurface FAILED");

        ZeroMemory(&ddscaps, sizeof(ddscaps));
        ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
        hRet = g_pDDSPrimary->GetAttachedSurface(&ddscaps, &g_pDDSBack);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "GetAttachedSurface FAILED");
		// Get the ever-important pitch information.
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		hRet = g_pDDSPrimary->GetSurfaceDesc(&ddsd);
		if (hRet != DD_OK) {
			return FALSE;
		}
		g_ddrawWidth = ddsd.dwWidth;
		g_ddrawHeight = ddsd.dwHeight;
		g_ddrawPitch = ddsd.lPitch / sizeof(Color);
		g_ddrawDepth = depth;
    }
    g_ready = TRUE;
    return DD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  When we decide to switch video modes, in WIN32, this is what we call.
Bool View::SetMode(u16 width, u16 height, u8 depth, Bool fullscreen) {
	fullscreen = fullscreen ? TRUE : FALSE;
	if ((width == g_ddrawWidth) &&
		(height == g_ddrawHeight) &&
		(depth == g_ddrawDepth) &&
		(fullscreen == g_fullscreen)) {
		return TRUE;
	}
	if (g_active && g_ready && (fullscreen != g_fullscreen)) {
		g_ready = FALSE;
		if (g_fullscreen) {
			GetWindowRect(g_hWnd, &g_rcWindow);
		}
	}
	#if !defined(DUMMY_BUFFER)
		HRESULT hRet;
		DDPIXELFORMAT ddpf;
		color c;

	    ReleaseMostObjects(g_hWnd);

		if (!g_pDD) {
			hRet = DirectDrawCreateEx(NULL, (VOID**)&g_pDD, IID_IDirectDraw7, NULL);
			if (hRet != DD_OK)
				InitFail(g_hWnd, hRet, "DirectDrawCreateEx FAILED");
		
			#if 0
				// Get exclusive mode
				hRet = g_pDD->SetCooperativeLevel(g_hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
				if (hRet != DD_OK)
					InitFail(g_hWnd, hRet, "SetCooperativeLevel FAILED");
			#endif
		}
		#if 1
		    if (!fullscreen) {
		        SetWindowPos(g_hWnd,
		        			 HWND_NOTOPMOST,
		        			 g_rcWindow.left,
		        			 g_rcWindow.top,
		                     (g_rcWindow.right - g_rcWindow.left), 
		                     (g_rcWindow.bottom - g_rcWindow.top),
		                     SWP_SHOWWINDOW);
		    }
		    hRet = InitSurfaces(g_hWnd, !fullscreen, width, height, depth);
		    if (hRet != DD_OK) {
		    	return FALSE;
		    }
		#else
			// Destroy Primary surface
			if (g_pDDSPrimary) {
			    g_pDDSPrimary->Release();
			    g_pDDSPrimary = NULL;
			}
			// Switch video mode
			hRet = g_pDD->SetDisplayMode(width, height, 8 * depth, 0, 0);
			if (hRet != DD_OK) {
				return FALSE;
			}
			
			// First, create complex flipping primary surface
			// Fill out surface description
			ZeroMemory(&ddsd, sizeof(ddsd));
			ddsd.dwSize = sizeof(ddsd);
			ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
			ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
								  DDSCAPS_FLIP |
								  DDSCAPS_COMPLEX;
			ddsd.dwBackBufferCount = 1;
		
			// Create the primary surface with 1 back buffer
			hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSPrimary, NULL);
			if (hRet != DD_OK) {
				return FALSE;
			}
		
			// Get pointer to back buffer
			ZeroMemory(&ddscaps, sizeof(ddscaps));
			ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
			hRet = g_pDDSPrimary->GetAttachedSurface(&ddscaps,
													 &g_pDDSBack);
			if (hRet != DD_OK) {
				return FALSE;
			}
			// Get the ever-important pitch information.
			ZeroMemory(&ddsd, sizeof(ddsd));
			ddsd.dwSize = sizeof(ddsd);
			hRet = g_pDDSPrimary->GetSurfaceDesc(&ddsd);
			if (hRet != DD_OK) {
				return FALSE;
			}
			g_ddrawWidth = ddsd.dwWidth;
			g_ddrawHeight = ddsd.dwHeight;
			g_ddrawPitch = ddsd.lPitch / sizeof(Color);
			g_ddrawDepth = depth;
		
			// Get information needed for color formation.
			ZeroMemory(&ddpf, sizeof(ddpf));
			ddpf.dwSize = sizeof(ddpf);
			ddpf.dwFlags = DDPF_RGB;
			hRet = g_pDDSPrimary->GetPixelFormat(&ddpf);
			if (hRet != DD_OK) {
				return FALSE;
			}
		#endif
	
		// Get information needed for color formation.
		ZeroMemory(&ddpf, sizeof(ddpf));
		ddpf.dwSize = sizeof(ddpf);
		ddpf.dwFlags = DDPF_RGB;
		hRet = g_pDDSPrimary->GetPixelFormat(&ddpf);
		if (hRet != DD_OK) {
			return FALSE;
		}

		g_redShift = 0;
		g_greenShift = 0;
		g_blueShift = 0;
		g_redMask = 0;
		g_greenMask = 0;
		g_blueMask = 0;
		if (ddpf.dwRBitMask) {
			for (c=ddpf.dwRBitMask; !(c & 1); c>>=1) {
				++g_redShift;
			}
			g_redMask = c;
		} else {
			return FALSE;
		}
	
		if (ddpf.dwGBitMask) {
			for (c=ddpf.dwGBitMask; !(c & 1); c>>=1) {
				++g_greenShift;
			}
			g_greenMask = c;
		} else {
			return FALSE;
		}
	
		if (ddpf.dwBBitMask) {
			for (c=ddpf.dwBBitMask; !(c & 1); c>>=1) {
				++g_blueShift;
			}
			g_blueMask = c;
		} else {
			return FALSE;
		}
	#else
		g_ddrawWidth = width;
		g_ddrawHeight = height;
		g_ddrawDepth = depth;
		g_redShift = 10;
		g_greenShift = 5;
		g_blueShift = 0;
		g_redMask = 0x1f;
		g_greenMask = 0x1f;
		g_blueMask = 0x1f;
		g_ddrawPitch = g_ddrawWidth;
	#endif
	g_fullscreen = fullscreen;
	Initialize();
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Make an "active graphic".  This is essentially an object whose visibility
// is managed by itself.
ActiveGraphic::ActiveGraphic(void) {
	if (m_next = s_allActiveGraphics) {
		m_next->m_prev = this;
	}
	m_prev = NULL;
	s_allActiveGraphics = this;
}

/////////////////////////////////////////////////////////////////////////////
//  Destroy the active graphic.
ActiveGraphic::~ActiveGraphic(void) {
	if (m_next) {
		m_next->m_prev = m_prev;
	}
	if (m_prev) {
		m_prev->m_next = m_next;
	} else {
		s_allActiveGraphics = m_next;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Redraw all active graphics.
void ActiveGraphic::RedrawAll(void) {
	ActiveGraphic * a;

	for (a=s_allActiveGraphics; a; a=a->m_next) {
		a->MakeDirty();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Make this graphic dirty.
void ActiveGraphic::MakeDirty(void) {
	m_needToUpdate = 0xff;
}

/////////////////////////////////////////////////////////////////////////////
//  Is this item dirty?
Bool ActiveGraphic::IsDirty(void) {
	u8 u = (1 << g_monitor.GetCurrentBuffer());

	return m_needToUpdate & u;
}

/////////////////////////////////////////////////////////////////////////////
//  Indicate that this frame has been updated.
void ActiveGraphic::CleanFrame(void) {
	u8 u = (1 << g_monitor.GetCurrentBuffer());

	m_needToUpdate &= ~u;
}
