/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere
	Copyright 1996, 1997 John Marco Panettiere

	Unified class for any 2D buffer that can be read from, written to, and
recognizes the notion of color.
	HERE how should the locking work?  I think we should manage it internally
so that the apps don't need to know whether the indicated feature is
supported in hardware or software.  However, we may have to add queueing
logic to maximize parallelism, or maybe a thread yield. . .
  	
****************************************************************************/

#ifndef VIEW_H
	#define VIEW_H

	class View : public Target < Color > {
		public:
			View(void);
			virtual ~View(void);

			u16 GetPitchInColor(void) const;	// this times sizeof(color) = pitch.
			Color * GetColorBuffer(void);		// for low-level renderers.
			u8 GetDepth(void) const;

			virtual void Refresh(void);			// Flip pages.
			virtual void Clear(Color u, Rect * p = NULL);	// Clear the screen.
			virtual Color MapColor(real, real, real) const;
			virtual void MapColor(Color c, real & r, real & g, real & b) const;
										// May use the Map method, but this
										// lets you address a pixel.
			// Import a BMP file and convert it to my color format.
			void ImportFile(const FString & filename,
							Map < Color > & m) const;
			// Export a BMP file that uses my color format.
			void ExportFile(const FString & filename,
							Map < Color > & m) const;
			virtual void Open(u8 option);
			virtual void Close(void);

			// This is a platform-specific member.
			#if defined(WIN32)
				static Bool SetMode(u16 width,
									u16 height,
									u8 depth,
									Bool fullscreen = TRUE);
			#elif defined(DOS)
				static Bool SetMode(u16 width,
									u16 height,
									u8 depth);
			#endif
		protected:
		private:
			void OLLock(void);			// Non-inlined portion.
			void OLUnlock(void);		// Non-inlined portion.

			static void Initialize(void);	// To init the monitor.
			Bool m_locked;
	};

	class ActiveGraphic {
		public:
			ActiveGraphic(void);
			virtual ~ActiveGraphic(void);

			static void RedrawAll(void);

			virtual void MakeDirty(void);
			virtual Bool IsDirty(void);
			virtual void CleanFrame(void);
		protected:
			u8 m_needToUpdate;
		private:
			ActiveGraphic * m_next, * m_prev;
			static ActiveGraphic * s_allActiveGraphics;
	};

	#include "view.ipp"
	#if defined(DOS)
		#include "viewdos.ipp"
	#endif

	extern View g_monitor;

#endif
