/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere
	Copyright 1996, 1997 John Marco Panettiere

	This the DOS implementation of my VIEW class.

****************************************************************************/

#include "all.h"

#define MAXMODES		128

#define X86REGS REGPACK
#define INTERRUPT(a, b) { if ((a)==0x10) rmi((a), (b)); else intr((a), (b)); }
#define MEM_INFO RealModeMemoryInfo

struct RealModeMemoryInfo {
	U32 size;
	U32 address;
};

#define WORD_REGS(a, b) (a).w.b
#define DWORD_REGS(a, b) (a).x.b
#define HIBYTE_REGS(a, b) (a).h.b
#define LOBYTE_REGS(a, b) (a).h.b

#define CLEAR_REGS(r) memset((r), 0, sizeof(*(r)))

/////////////////////////////////////////////////////////////////////////////
//  A single palette item.
#pragma pack(1)
struct VBE2Pal {
	U8		blue;
	U8		green;
	U8		red;
	U8		reserved;
};

#pragma pack(1)
/////////////////////////////////////////////////////////////////////////////
//  Information about installed VESA.
struct VBE2Inf {
	char	VbeSignature[4];
	U16		VbeVersion;
	U32 	OemStringPtr;
	U32		Capabilities;
	U16 	VideoModeOff;
	U16		VideoModeSeg;
	U16		TotalMemory;
	U16		OemSoftwareRev;
	U32		OemVendorNamePtr;
	U32		OemProductNamePtr;
	U32		OemProductRevPtr;
	U8		Reserved[222];
	U8		OemData[256];
};

#pragma pack(1)
/////////////////////////////////////////////////////////////////////////////
//  Information about a given mode.
struct VBE2Mod {
	U16		ModeAttributes;
	U8		WinAAttributes;
	U8		WinBAttributes;
	U16		WinGranularity;
	U16		WinSize;
	U16		WinASegment;
	U16		WinBSegment;
	U32		WinFuncPtr;
	U16		BytesPerScanline;
	U16		XResolution;
	U16		YResolution;
	U8		XCharSize;
	U8		YCharSize;
	U8		NumberOfPlanes;
	U8		BitsPerPixel;
	U8		NumberOfBanks;
	U8		MemoryModel;
	U8		BankSize;
	U8		NumberOfImagePages;
	U8		Reserved1;
	
	U8		RedMaskSize;
	U8		RedFieldPosition;
	U8		GreenMaskSize;
	U8		GreenFieldPosition;
	U8		BlueMaskSize;
	U8		BlueFieldPosition;
	U8		RsvdMaskSize;
	U8		RsvdFieldPosition;
	U8		DirectColorModeInfo;
	U32		PhysBasePtr;
	U32		OffScreenMemOffset;
	U16		OffScreenMemSize;
	U8		Reserved2[206];
};

/////////////////////////////////////////////////////////////////////////////
//  Our own information about each mode.  In the future, direct color info
// should also be provided, but for the video game, it won't be needed.
struct DModes {
	U32		width;
	U32		height;
	U32		pitch;
	U32		frameBuffer;
	U32		modeNumber;					// VESA mode, that is. . .
	U32		redMask;
	U32		greenMask;
	U32		blueMask;
	U32		redPosition;
	U32		greenPosition;
	U32		bluePosition;	
	U8		bpp;
	U8		pages;
};

VBE2Pal * g_palette = NULL;
DModes * g_modeInfo = NULL;
VBE2Pal	 	g_vbe2Palette;						// Palette.
VBE2Inf		g_vbe2Vbeinfo;						// VBE information.
DModes		g_drawModes[MAXMODES];				// Our mode information.

U16		g_rmSelector=0;
U16		g_rmSegment;
U32		g_vbeInitialized = FALSE;	// Check before using vbe.
U8		g_numVidmodes;			// How many modes we support.
U32		g_videoNumberOfLinearPages;		// How many buffers we have.
U8 *	g_videoPages[3];		// The display buffers.
U32		g_activePage;			// The current display page.

u16 g_dosWidth;					// Horizontal resolution.
u16 g_dosHeight;			// Vertical resolution.
static U32 s_videoPitch;			// Actual #bytes across.
U32		g_dosDepth;				// Bytes per pixel.
MType	g_dosPitch;				// Colors per pixel.
U32		g_vidmemAvailable;		// How much available memory.
U8 *	g_baseVideoMemory;		// Linear framebuffer, what we're after.

U32		g_masks[9] = { 0x0, 0x1, 0x3, 0x7, 0xf, 0x1f, 0x3f, 0x7f, 0xff };

static U32 g_tb;
static U32 s_selector;
static U8 far * s_allocatedMemory;

#define DSIZE Max((256 * sizeof(VBE2Pal)), Max(sizeof(VBE2Inf), sizeof(VBE2Mod)))

static void rmi(U32, X86REGS *);

#if defined(VIDEO_MODE_ANALYSIS)
	static FILE * s_vidReport;
#endif

static void InitializeSegment(void) {
	X86REGS regs;

	CLEAR_REGS(&regs);

	WORD_REGS(regs, ax) = 0x100;
	WORD_REGS(regs, bx) = (DSIZE + 15) / 16;
	INTERRUPT(0x31, &regs);

	//	Paragraph of memory allocated.
	g_tb = (U32)WORD_REGS(regs, ax) * 16;
	s_selector = WORD_REGS(regs, dx);

	s_allocatedMemory = (U8 far *)(MK_FP(s_selector, 0));
}

static void dosmemget(int ignored,
					  U32 size,
					  void * target) {
	U8 * dest = (U8 *)target;

	while (size--) {
		dest[size] = s_allocatedMemory[size];
	}
	ignored = ignored;
}

static void dosmemput(void * src,
					  U32 size,
					  int ignored) {
	U8 * source = (U8 *)src;

	while (size--) {
		s_allocatedMemory[size] = source[size];
	}
	ignored = ignored;
}

#pragma pack(1)
struct RealModeRegs {
	U32 edi;
	U32 esi;
	U32 ebp;
	U32 reserved;
	U32 ebx;
	U32 edx;
	U32 ecx;
	U32 eax;
	U16 flags;
	U16 es;
	U16 ds;
	U16 fs;
	U16 gs;
	U16 ip;
	U16 cs;
	U16 sp;
	U16 ss;
};

/////////////////////////////////////////////////////////////////////////
//  Effect a real-mode interrupt.  I wonder why CWSDPMI doesn't have a
// problem with this?
static void rmi(U32 inumber,
				X86REGS * retval) {
	RealModeRegs s_rmregs;
	X86REGS regs;

	memset(&s_rmregs,0,sizeof(s_rmregs));
	s_rmregs.eax=DWORD_REGS(*retval, eax);
	s_rmregs.ebx=DWORD_REGS(*retval, ebx);
	s_rmregs.ecx=DWORD_REGS(*retval, ecx);
	s_rmregs.edx=DWORD_REGS(*retval, edx);
	s_rmregs.esi=DWORD_REGS(*retval, esi);
	s_rmregs.edi=DWORD_REGS(*retval, edi);
	s_rmregs.ds =WORD_REGS(*retval, ds);
	s_rmregs.es =WORD_REGS(*retval, es);
	s_rmregs.flags=0;

	CLEAR_REGS(&regs);

	DWORD_REGS(regs, eax) = 0x0300;
	DWORD_REGS(regs, ebx) = inumber;
	DWORD_REGS(regs, ecx) = 0;
	DWORD_REGS(regs, edi) = FP_OFF(&s_rmregs);
	WORD_REGS(regs, es)   = FP_SEG(&s_rmregs);

	INTERRUPT(0x31, &regs);

	DWORD_REGS(*retval, eax) = s_rmregs.eax;
	DWORD_REGS(*retval, ebx) = s_rmregs.ebx;
	DWORD_REGS(*retval, ecx) = s_rmregs.ecx;
	DWORD_REGS(*retval, edx) = s_rmregs.edx;
	DWORD_REGS(*retval, esi) = s_rmregs.esi;
	DWORD_REGS(*retval, edi) = s_rmregs.edi;
}

static U8 far * __dpmi_map_real_mode_segment(u16 segment) {
	X86REGS regs;

	CLEAR_REGS(&regs);

	WORD_REGS(regs, ax) = 0x2;
	WORD_REGS(regs, bx) = segment;

	INTERRUPT(0x31, &regs);

	if (regs.x.flags & INTR_CF)
		return (U8 far *)NULL;

	return (U8 far *)(MK_FP(WORD_REGS(regs, ax), 0));
}

static int __dpmi_physical_address_mapping(RealModeMemoryInfo * r) {
	X86REGS regs;

	CLEAR_REGS(&regs);

	WORD_REGS(regs, ax) = 0x800;
	WORD_REGS(regs, bx) = r->address >> 16;
	WORD_REGS(regs, cx) = r->address & 0xffff;
	WORD_REGS(regs, si) = r->size >> 16;
	WORD_REGS(regs, di) = r->size & 0xffff;

	INTERRUPT(0x31, &regs);

	if (regs.x.flags & INTR_CF)
		return -1;

	r->address = ((U32)WORD_REGS(regs, bx) << 16) |
				  WORD_REGS(regs, cx);
	return 0;
}

#define __djgpp_conventional_base				0

/////////////////////////////////////////////////////////////////////////////
//  This function tells us if the given videomode is one that we want to use
// or not.  If it is, we fill in our own video mode information structure
// and return "TRUE", otherwise we return "FALSE".
static U32 GoodMode(VBE2Mod * vm, DModes * dm, U32 mode) {
	if ((vm->ModeAttributes & 1) &&
		(vm->ModeAttributes & 8) &&			// Color mode.
		(vm->ModeAttributes & 16) &&		// Graphics mode.
		(vm->ModeAttributes & 128) &&		// Linear framebuffer available.
		(vm->NumberOfImagePages >= BUFFER_COUNT)) {	// Double buffering.
		dm->width			= vm->XResolution;
		dm->height			= vm->YResolution;
		dm->bpp				= vm->BitsPerPixel >> 3;
		dm->modeNumber		= mode;
		dm->pitch			= vm->BytesPerScanline;
		dm->pages			= vm->NumberOfImagePages;
		dm->frameBuffer		= (U32)vm->PhysBasePtr;
		dm->redMask			= g_masks[vm->RedMaskSize];
		dm->greenMask		= g_masks[vm->GreenMaskSize];
		dm->blueMask		= g_masks[vm->BlueMaskSize];
		dm->redPosition		= vm->RedFieldPosition;
		dm->greenPosition	= vm->GreenFieldPosition;
		dm->bluePosition	= vm->BlueFieldPosition;

		#if defined(VIDEO_MODE_ANALYSIS)
			fprintf(s_vidReport, "Mode found - \n");
			fprintf(s_vidReport, "\t%x mode.\n", (unsigned)dm->modeNumber);
			fprintf(s_vidReport, "\t%x width.\n", (unsigned)dm->width);
			fprintf(s_vidReport, "\t%x height.\n", (unsigned)dm->height);
			fprintf(s_vidReport, "\t%x bits per pixel.\n", (unsigned)dm->bpp);
			fprintf(s_vidReport, "\t%x pitch.\n", (unsigned)dm->pitch);
			fprintf(s_vidReport, "\t%x pages.\n\n", (unsigned)dm->pages);
		#endif
	}
	else {
		return FALSE;
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  This is how we get information about a given mode.
static Bool GetVbe2ModeInformation(U32 mode, VBE2Mod * mi) {
	X86REGS regs;

	CLEAR_REGS(&regs);

	WORD_REGS(regs,ax) = 0x4f01;
	WORD_REGS(regs,cx) = mode;
	WORD_REGS(regs,es) = g_tb / 16;
	WORD_REGS(regs,di) = 0;
	INTERRUPT(0x10, &regs);

	if (HIBYTE_REGS(regs,ah)) {
		return FALSE;
	}
	
	dosmemget(g_tb, sizeof(VBE2Mod), mi);
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Our first function:  check for vbe2 being present.
static U32 Vbe2Open(void) {
	X86REGS regs;
	VBE2Mod vbe2mi;
	U32 i;

	CLEAR_REGS(&regs);

	if (g_vbeInitialized) {
		return TRUE;
	}

	InitializeSegment();

	strncpy(g_vbe2Vbeinfo.VbeSignature, "VBE2", 4);
	dosmemput(&g_vbe2Vbeinfo, sizeof(VBE2Inf), g_tb);
	WORD_REGS(regs,ax) = 0x4f00;
	WORD_REGS(regs,di) = 0;
	WORD_REGS(regs,es) = g_tb / 16;
	INTERRUPT(0x10, &regs);
	dosmemget(g_tb, sizeof(VBE2Inf), &g_vbe2Vbeinfo);
	
// Splendid.  I've just retrieved the information I need.  That wasn't so
// bad, was it?  Now, let me get all mode that I can use.
	if (strncmp(g_vbe2Vbeinfo.VbeSignature,"VESA",4)) {
		return FALSE;
	}
	
// Is it the version we want?
	if (g_vbe2Vbeinfo.VbeVersion < 0x200) {
		return FALSE;
	}

	#if defined(VIDEO_MODE_ANALYSIS)
		s_vidReport = fopen("vids.txt", "w");
	#endif

// Okay, we have to cycle through all the available modes to see which ones
// are fitting for our purposes.  How do we get the array of video modes?
// I will try some of the nearpointer tricks!
	#if 0
		if (!(mode = (S16 far *)__dpmi_map_real_mode_segment(g_vbe2Vbeinfo.VideoModeSeg)))
			return FALSE;
		mode += g_vbe2Vbeinfo.VideoModeOff >> 1;
		for (i=0, g_numVidmodes=0; i<111; i++, mode++) {
			if (*mode == -1)
				break;
			if (GetVbe2ModeInformation(*mode, &vbe2mi))
				if (GoodMode(&vbe2mi, g_drawModes + g_numVidmodes, *mode))
					g_numVidmodes++;
		}
	#else
		// HERE we have to convert this to a config file system.
		FILE * vm;

		if (!(vm = fopen("vid.cfg", "rb"))) {
			for (i=0, g_numVidmodes=0;
				 (i<65536) && (g_numVidmodes < 111);
				 i++) {
				if (GetVbe2ModeInformation(i, &vbe2mi))
					if (GoodMode(&vbe2mi, g_drawModes + g_numVidmodes, i))
						g_numVidmodes++;
			}
			if (vm = fopen("vid.cfg", "wb")) {
				fwrite(&g_numVidmodes, sizeof(g_numVidmodes), 1, vm);
				fwrite(g_drawModes, sizeof(*g_drawModes), g_numVidmodes, vm);
				fclose(vm);
			}
		} else {
			fread(&g_numVidmodes, sizeof(g_numVidmodes), 1, vm);
			fread(g_drawModes, sizeof(*g_drawModes), g_numVidmodes, vm);
			fclose(vm);
		}
	#endif

	#if defined(VIDEO_MODE_ANALYSIS)
		fclose(s_vidReport);
	#endif
// Very, very cool.  We've determined if vbe is available, and if so, we've
// checked which modes are usable.  Now, if at least one mode can be used,
// call me a starry-eyed optimist, but I'm reporting success!
	if (g_numVidmodes == 0)
		return FALSE;
		
	g_vidmemAvailable = g_vbe2Vbeinfo.TotalMemory * 65536;
	g_vbeInitialized = TRUE;
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  If we cannot set vbe2 modes, at least we have our old friend.
static U8 * SetMode13h(void) {
	MEM_INFO info;
	X86REGS regs;

	CLEAR_REGS(&regs);

	info.size		= 65536;
	info.address	= 0xa0000;

// Before allocating the memory, let's try setting the mode.	
	WORD_REGS(regs,ax)=0x13;
	INTERRUPT(0x10,&regs);
	
	if(__dpmi_physical_address_mapping(&info) == -1) {
		return NULL;
	}

// Now, we want a near pointer to that address that dpmi just gave us...
	g_baseVideoMemory=(U8 * )(info.address + __djgpp_conventional_base);
	
	return g_baseVideoMemory;
}

/////////////////////////////////////////////////////////////////////////////
//  Now, our second function:  Set the desired mode.  For now, an exact match
// must be found.  If vbe is not present, we set mode 0x13.
static U8 * SetDesiredMode(u16 & desired_width,
						   u16 & desired_height,
						   MType & desired_pitch,
						   MType & desired_bpp,
						   DModes ** dm) {
	U32 i;
	U8 * retval;
	MEM_INFO info;
	X86REGS regs;

	CLEAR_REGS(&regs);

	if (!g_vbeInitialized) {
		if (!Vbe2Open()) {
			g_dosWidth = desired_width = 320;
			g_dosHeight = desired_height = 200;
			g_dosDepth = desired_bpp = 1;
			s_videoPitch = desired_pitch = 320;
			desired_pitch /= sizeof(Color);

			g_videoNumberOfLinearPages=1;
			g_videoPages[0] = new U8[s_videoPitch*g_dosHeight];
			g_videoPages[1] = new U8[s_videoPitch*g_dosHeight];
			g_activePage=0;
			
			return SetMode13h();
		}
	}
		
// Must match _everything_ now, in the future we will best fit.
	for (i=0; i<g_numVidmodes;i++) {
		if ((g_drawModes[i].width == desired_width) &&
			(g_drawModes[i].height == desired_height) &&
			(g_drawModes[i].bpp == desired_bpp))
			break;
	}
	
	if (i == g_numVidmodes)
		return NULL;
		
// Great.  Let's set the mode.
	g_dosWidth = desired_width 	=	g_drawModes[i].width;
	g_dosHeight = desired_height	=	g_drawModes[i].height;
	g_dosDepth = desired_bpp		=	g_drawModes[i].bpp;
	s_videoPitch = desired_pitch	=	g_drawModes[i].pitch;
	desired_pitch /= sizeof(Color);
	
	*dm = g_drawModes + i;

	info.size		= g_vidmemAvailable;
	info.address	= g_drawModes[i].frameBuffer;

// Before allocating the memory, let's try setting the mode.	
	WORD_REGS(regs,ax)=0x4f02;
	WORD_REGS(regs,bx)=0x4000 | g_drawModes[i].modeNumber;	// Use linear.
	INTERRUPT(0x10,&regs);
	
	if ((LOBYTE_REGS(regs,al) != 0x4f) || HIBYTE_REGS(regs,ah)) {
		return NULL;
	}
	
	if(__dpmi_physical_address_mapping(&info) == -1) {
		return NULL;
	}

// Now, we want a near pointer to that address that dpmi just gave us...
	g_baseVideoMemory=(U8 * )(info.address + __djgpp_conventional_base);

// HERE:  Additional work needs to be done to properly accomodate virtual
// page sizes.
	g_videoNumberOfLinearPages=2;
	g_videoPages[0] = g_baseVideoMemory;
	g_videoPages[1] = g_baseVideoMemory + g_dosHeight*s_videoPitch;
	g_activePage=0;
	
	retval = g_baseVideoMemory;
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  This function changes the active video page for video modes with display
// pages in display memory.  It actually changes active scanline.  Future
// revisions might require me to set column, or permit me to (which is
// probably better).
static void SetVisiblePage(U32 desiredRow,
						   U32 desiredColumn) {
	X86REGS regs;

	CLEAR_REGS(&regs);

	WORD_REGS(regs,ax) = 0x4f07;
	WORD_REGS(regs,bx) = 0x80;
	WORD_REGS(regs,cx) = desiredColumn;
	WORD_REGS(regs,dx) = desiredRow;
	INTERRUPT(0x10,&regs);
}

/////////////////////////////////////////////////////////////////////////////
//  This function blits a page.  Right now, banked video modes are not
// accounted for, nor are hardware accelerators considered.
void BlitPage(U8 * p) {
	memcpy(g_baseVideoMemory, p, s_videoPitch * g_dosHeight);

}

/////////////////////////////////////////////////////////////////////////////
//  Initialize everything to nothing.
View::View(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  Initialize the dosview.
void View::Initialize(void) {
	MType m;
	Color * c;

	g_monitor.m_elements.Make(g_dosHeight);
	g_monitor.m_width = g_dosWidth;
	for (m=0; m<BUFFER_COUNT; ++m) {
		g_monitor.m_overlays[m].Destroy();
	}
	g_activePage = 0;

	#if !defined(DUMMY_BUFFER)
		SetDesiredMode(g_dosWidth,
					   g_dosHeight,
					   g_dosPitch,
					   g_dosDepth,
					   &g_modeInfo);
		for (m=0, c=(Color *)g_videoPages[0];
			 m<g_dosHeight;
			 ++m, c+=g_dosPitch) {
			g_monitor.m_elements[m] = c;
		}
	#else
		g_modeInfo = new DModes;
		g_dosPitch = g_dosWidth;
		g_videoPages[0] = new U8[g_dosPitch * sizeof(Color) * g_dosHeight];

		g_modeInfo->width = g_dosWidth;
		g_modeInfo->height = g_dosHeight;
		g_modeInfo->pitch = g_dosPitch * sizeof(Color);
		g_modeInfo->frameBuffer = (U32)g_videoPages[0];
		g_modeInfo->modeNumber = 0xffffffff;
		g_modeInfo->redMask = 0x1f;
		g_modeInfo->greenMask = 0x1f;
		g_modeInfo->blueMask = 0x1f;
		g_modeInfo->redPosition = 10;
		g_modeInfo->greenPosition = 5;
		g_modeInfo->bluePosition = 0;
		g_modeInfo->bpp = 16;
		g_modeInfo->pages = 1;
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Page-flipping.
void View::Refresh(void) {
	static u16 s_currentRow;
	MType m;
	Color * c;

	Flush();
	#if !defined(DUMMY_BUFFER)
		// HEREHEREHERE a separate option for blitted pages:
		// substitute BlitPage(g_videoPage[g_activePage]); for below.
		SetVisiblePage(s_currentRow, 0);
		++g_activePage;
		s_currentRow += g_dosHeight;
		if (g_activePage == BUFFER_COUNT) {
			s_currentRow = 0;
			g_activePage = 0;
		}
		// HEREHEREHERE a pointer switch would be faster. . .
		for (m=0, c=(Color *)g_videoPages[g_activePage];
			 m<g_dosHeight;
			 ++m, c+=g_dosPitch) {
			g_monitor.m_elements[m] = c;
		}
	#endif
	UndoOverlayDraw();
}

/////////////////////////////////////////////////////////////////////////////
//  This destructor is needed to restore video mode.
View::~View(void) {
	#if !defined(DUMMY_BUFFER)
		X86REGS regs;

		CLEAR_REGS(&regs);

		WORD_REGS(regs,ax) = 3;
		INTERRUPT(0x10, &regs);
		m_elements[0] = NULL;	// So receptor destructor doesn't free it.
	#else
		if (g_videoPages[0]) {
			delete [] g_videoPages[0];
		}
	#endif
}

#if 0
/////////////////////////////////////////////////////////////////////////////
//  This private function is used to get the palette.  We hide the fact that
// a palette is used, should we?  It wouldn't hurt to do so for now, we can
// expose it later if we want to.
//	I think I'm going to have to enable palette-getting in the future.
// That's not a problem, modes that need it can use it, and modes that don't
// can ignore it.
void View::SetPalette(u32 index, Real red, Real green, Real blue) {
	if (m_mode == SCREEN) {
		X86REGS regs;

		CLEAR_REGS(&regs);

		if (g_dosDepth > 1)
			return;
		if (!g_palette) {
			g_palette=new VBE2Pal[256];
			memset(g_palette, 0, 256*sizeof(VBE2Pal));
		}
		g_palette[index].reserved	= 0;
		g_palette[index].red		= (u32)(red * (Real)63.0);
		g_palette[index].green		= (u32)(green * (Real)63.0);
		g_palette[index].blue		= (u32)(blue * (Real)63.0);
		
	//	If in vesa mode, use vesagetpalette; else use vga get palette.  Damn!
	// We have to use that "transfer buffer" crap!  Note:  we also, in the
	// future, have to accomodate those different shift/mask combinations.
	//  For now, 6-bit rgb is assumed.
		if (index == 255) {
			if (g_vbeInitialized) {
				dosmemput(g_palette, 256 * sizeof(VBE2Pal), g_tb);
				
				WORD_REGS(regs,ax) = 0x4f09;
				WORD_REGS(regs,bx) = 0x80;
				WORD_REGS(regs,cx) = 256;
				WORD_REGS(regs,dx) = 0;
				WORD_REGS(regs,di) = 0;
				WORD_REGS(regs,es) = g_tb / 16;
				INTERRUPT(0x10, &regs);
			} else {
				WORD_REGS(regs,ax) = 0x1012;
				WORD_REGS(regs,bx) = 0;
				WORD_REGS(regs,cx) = 256;
				WORD_REGS(regs,dx) = 0;
				WORD_REGS(regs,es) = g_tb / 16;
				
		//	DEBUG:  The palette format is different under VGA, this should be
		// accounted for.
				dosmemput(g_palette, 256 * sizeof(VBE2Pal), g_tb);
				
				INTERRUPT(0x10, &regs);
			}
		}
	}
}
#endif

/////////////////////////////////////////////////////////////////////////////
//  This function maps a pixel to a color, based on 3 floating point values
// passed to us.  The floats are a more generic representation of the data.
Color View::MapColor(real r, real g, real b) const {
	s32 red, green, blue;
	s32 u, best, vv, w;
	U8 bestu=0;

	if (g_dosDepth == 1) {
		red = (u32)(r * (Real)63.0);		// DEBUG!  6-bit hues assumed.
		green = (u32)(g * (Real)63.0);
		blue = (u32)(b * (Real)63.0);
		
		for (u=0, best=16777217; u<255; u++) {
			vv=g_palette[u].red;
			vv -= (s32)red;
			w = vv*vv;
			vv=g_palette[u].green;
			vv -= (s32)green;
			w += vv*vv;
			vv=g_palette[u].blue;
			vv -= (s32)blue;
			w += vv*vv;
			if (w < best) {
				best=w;
				bestu=u;
			}
		}
		return bestu;
	} else if (g_dosDepth == 2) {
		u = (u32)(r * (Real)(g_modeInfo->redMask));
		u = u & g_modeInfo->redMask;
		u = u << g_modeInfo->redPosition;
		best=u;
		u = (u32)(g * (Real)(g_modeInfo->greenMask));
		u = u & g_modeInfo->greenMask;
		u = u << g_modeInfo->greenPosition;
		best |= u;
		u = (u32)(b * (Real)(g_modeInfo->blueMask));
		u = u & g_modeInfo->blueMask;
		u = u << g_modeInfo->bluePosition;
		best |= u;
		return best;
	} else if (g_dosDepth == 3) {
		u = (u32)(r * (Real)(g_modeInfo->redMask));
		u = u & g_modeInfo->redMask;
		u = u << g_modeInfo->redPosition;
		best=u;
		u = (u32)(g * (Real)(g_modeInfo->greenMask));
		u = u & g_modeInfo->greenMask;
		u = u << g_modeInfo->greenPosition;
		best |= u;
		u = (u32)(b * (Real)(g_modeInfo->blueMask));
		u = u & g_modeInfo->blueMask;
		u = u << g_modeInfo->bluePosition;
		best |= u;
		return best;
	} else {
		return 0;
	}
}

#if 0
/////////////////////////////////////////////////////////////////////////////
//  Okay, here we do the opposite.  We pass in a palette entry, and learn
// what its RGB values are.
void View::GetPalette(u32 p, Real & r, Real & g, Real & b) {
	if (g_dosDepth == 1) {
		r = ((Real)(g_palette[p].red)) / (Real)63.0;
		g = ((Real)(g_palette[p].green)) / (Real)63.0;
		b = ((Real)(g_palette[p].blue)) / (Real)63.0;
	} else {			// direct color mode.
		r = (Real)((p >> g_modeInfo->redPosition) & g_modeInfo->redMask);
		r /= (Real)(g_modeInfo->redMask);
		g = (Real)((p >> g_modeInfo->greenPosition) & g_modeInfo->greenMask);
		g /= (Real)(g_modeInfo->greenMask);
		b = (Real)((p >> g_modeInfo->bluePosition) & g_modeInfo->blueMask);
		b /= (Real)(g_modeInfo->blueMask);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This is a SLOW function that can be used for those situations where we
// don't need to worry about speed.  It allows us to quickly throw together
// other primitives.
void View::SetRGBPixel(u16 x,
					   u16 y,
					   Real red,
					   Real green,
					   Real blue) {
	U8 * buf;
	U32 putValue;
	U32 j;

// We clip here.
	if ((y >= g_dosHeight) || (x >= g_dosWidth))
		return;

// Saturate.
	if (red < (Real)0.0)
		red = (Real)0.0;
	else if (red >= (Real)1.0)
		red = (Real)0.9999;
	
	if (green < (Real)0.0)
		green = (Real)0.0;
	else if (green >= (Real)1.0)
		green = (Real)0.9999;
	
	if (blue < (Real)0.0)
		blue = (Real)0.0;
	else if (blue >= (Real)1.0)
		blue = (Real)0.9999;
	
	buf = Lock() + y*g_dosPitch + x*g_dosDepth;
	j = (u32)(red * ((Real)g_modeInfo->redMask));
	j &= g_modeInfo->redMask;
	putValue = j << g_modeInfo->redPosition;
	
	j = (u32)(green * ((Real)g_modeInfo->greenMask));
	j &= g_modeInfo->greenMask;
	putValue |= j << g_modeInfo->greenPosition;
	
	j = (u32)(blue * ((Real)g_modeInfo->blueMask));
	j &= g_modeInfo->blueMask;
	putValue |= j << g_modeInfo->bluePosition;
	
	for (j=0; j<g_dosDepth; j++) {
		*buf++ = putValue & 0xff;
		putValue >>= 8;
	}
	Unlock();
}
#endif

/////////////////////////////////////////////////////////////////////////////
//  Clear the indicated region to the indicated color.
void View::Clear(Color u, Rect * p) {
	Open(WRITE);
	if (p) {
		u16 x, y;

		for (y=0; y<p->h; ++y) {
			for (x=0; x<p->w; ++x) {
				m_elements[y + p->y][x + p->x] = u;
			}
		}
	} else {
		Color * c = (Color *)g_videoPages[g_activePage];
		MType m = g_dosPitch * g_dosHeight;

		while (m--) {
			c[m] = u;
		}
	}
	Close();
}

/////////////////////////////////////////////////////////////////////////////
//  Change the video mode on-the-fly.
void View::SetMode(u16 width, u16 height, u8 depth) {
	MType m;
	Color * c;

	if ((width == g_dosWidth) &&
		(height == g_dosHeight) &&
		(depth == g_dosDepth)) {
		return;
	}
	g_dosWidth = width;
	g_dosHeight = height;
	g_dosDepth = depth;
	
	#if !defined(DUMMY_BUFFER)
	#else
		g_dosPitch = g_dosWidth;
		delete [] g_videoPages[0];
		g_videoPages[0] = new U8[g_dosPitch * sizeof(Color) * g_dosHeight];
	#endif
	Initialize();
}

/////////////////////////////////////////////////////////////////////////////
//  Update the state of the mouse.
u8 MousePoll(void *) {
	while(1) {
		X86REGS regs;
	
		memset(&regs, 0, sizeof(regs));
		WORD_REGS(regs, ax) = 0xb;
		INTERRUPT(0x33, &regs);
		g_mouseX += (S16)(WORD_REGS(regs, cx));
		g_mouseY += (S16)(WORD_REGS(regs, dx));
	
		if ((s16)g_mouseX < 0)
			g_mouseX = 0;
		else if (g_mouseX >= g_monitor.GetWidth())
			g_mouseX = g_monitor.GetWidth() - 1;
	
		if ((s16)g_mouseY < 0)
			g_mouseY = 0;
		else if (g_mouseY >= g_monitor.GetHeight())
			g_mouseY = g_monitor.GetHeight() - 1;
	
		// Now, let's determine the button statuses.
		WORD_REGS(regs, ax) = 0x3;
		INTERRUPT(0x33, &regs);
		if (WORD_REGS(regs, bx) & 1)
			MouseHotspotGroup::Click(g_mouseX, g_mouseY);
		if (WORD_REGS(regs, bx) & 2)
			MouseHotspotGroup::RClick(g_mouseX, g_mouseY);
		ThreadYield();
	}
	return 0;
}
