/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of my windows.

****************************************************************************/

#include "all.h"

List < Window * > Window::s_allWindows;
FontHandle g_windowFont;

/////////////////////////////////////////////////////////////////////////////
//  Delete a window.
Window::~Window(void) {
	MType m;

	for (m=0; m<m_tabs.Count(); ++m) {
		SafeDelete(m_tabs[m]);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Set the dimensions of the window.
void Window::SetXYWH(u16 x, u16 y, u16 w, u16 h) {
	ClientSpot * p;
	u16 fh = TextGetFontHeight(g_windowFont);

	m_x = x;
	m_y = y;
	m_w = w;
	m_h = h;
	m_tabs.Destroy();
	SafeNew(p, (MType)1, __FILE__, __LINE__);
	p->SetArguments(this);
	p->SetDims(m_x + 2, m_y + fh + 2, m_w - 4, m_h - fh - 4);
	m_tabs += p;
}

/////////////////////////////////////////////////////////////////////////////
//  Need the virtual destructor (or may need it).
ClientSpot::~ClientSpot(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  This handles clicking in the window's client area.
void ClientSpot::Click(u16 x, u16 y) {
	if (InRegion(x, y)) {
		m_window->m_panels[m_window->m_selection].redraw->Click(x, y);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This handles right-clicking in the window's client area.
void ClientSpot::RClick(u16 x, u16 y) {
	if (InRegion(x, y)) {
		m_window->m_panels[m_window->m_selection].redraw->RClick(x, y);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This handles clicking on the tabs.
void WindowSpot::Click(u16 x, u16 y) {
	if (InRegion(x, y)) {
		m_window->SetTab(m_tab);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Add the indicated panel to the list.
void Window::AddPanel(const FString & name,
					  Panel * redraw) {
	WindowPanel p;
	WindowSpot * m;
	u16 x, y, w, h;
	u16 fw = TextGetFontWidth(g_windowFont);
	MType i;

	x = m_x;
	for (i=0; i<m_panels.Count(); ++i) {
		x += (m_panels[i].name.Length() + 2) * fw;
	}
	w = fw * (name.Length() + 2);
	y = m_y;
	h = TextGetFontHeight(g_windowFont);

	p.name = name;
	p.redraw = redraw;
	m_panels += p;

	if (redraw) {
 		redraw->SetDims(m_x + 2, m_y + h, m_w - 4, m_h - h - 2);
		redraw->SetParent(this);
 		redraw->GenerateContent();
	}

	SafeNew(m, (MType)1, __FILE__, __LINE__);
	m->SetDims(x, y, w, h);
	m->SetArguments(this, m_panels.Count() - 1);
	m_tabs += m;
}

/////////////////////////////////////////////////////////////////////////////
//  Update the window.
void Window::Refresh(void) {
	if (IsDirty()) {
		u16 fh = TextGetFontHeight(g_windowFont);
		u16 fw = TextGetFontWidth(g_windowFont);
		MType m;
		u16 x;

		// First, draw the border.
		DrawLine(g_monitor,
				 g_colors[BORDER],
				 m_x + 1,
				 m_y + fh - 1,
				 m_x + 1,
				 m_y + m_h - 1);
		DrawLine(g_monitor,
				 g_colors[BORDER],
				 m_x + 1,
				 m_y + m_h - 1,
				 m_x + m_w - 1,
				 m_y + m_h - 1);
		DrawLine(g_monitor,
				 g_colors[BORDER],
				 m_x + m_w - 1,
				 m_y + fh - 1,
				 m_x + m_w - 1,
				 m_y + m_h - 1);

		// Now, let's write the strings.
		x = m_x + 1;
		for (m=0; m<m_panels.Count(); ++m) {
			MType n;
			color c;
			FString f;

			for (n=0; n<m_panels[m].name.Length() + 2; ++n) {
				f += (char)0x13;
			}
			if (m != m_selection) {
				c = g_colors[UNSELECTED];
				TextWrite(g_monitor,
						  g_windowFont,
						  x,
						  m_y,
						  g_colors[BORDER],
						  f);
			} else {
				c = g_colors[SELECTED];
				TextWrite(g_monitor,
						  g_windowFont,
						  x,
						  m_y,
						  g_colors[BACKGROUND],
						  f);
			}
			TextWrite(g_monitor,
					  g_windowFont,
					  x,
					  m_y,
					  g_colors[BORDER],
					  (char)0x10,
					  Bitmap::CUTOUT | Bitmap::COLORIZE);
			x += fw;
			f.Destroy();
			for (n=0; n<m_panels[m].name.Length(); ++n) {
				f += (char)0x12;
			}
			TextWrite(g_monitor,
					  g_windowFont,
					  x,
					  m_y,
					  c,
					  m_panels[m].name,
					  Bitmap::CUTOUT | Bitmap::COLORIZE);
			TextWrite(g_monitor,
					  g_windowFont,
					  x,
					  m_y,
					  g_colors[BORDER],
					  f,
					  Bitmap::CUTOUT | Bitmap::COLORIZE);
			x += fw * m_panels[m].name.Length();
			TextWrite(g_monitor,
					  g_windowFont,
					  x,
					  m_y,
					  g_colors[BORDER],
					  (char)0x11,
					  Bitmap::CUTOUT | Bitmap::COLORIZE);
			x += fw;
		}

		// Now, let's draw a line from the end of the last tab to the end
		// of the window region.
		DrawLine(g_monitor,
				 g_colors[BORDER],
				 x,
				 m_y + fh - 1,
				 m_x + m_w - 1,
				 m_y + fh - 1);
		CleanFrame();
	}
	if (m_panels[m_selection].redraw)
		m_panels[m_selection].redraw->Redraw();
}

/////////////////////////////////////////////////////////////////////////////
//  When we've added everything to this window that we want, let's make it
// available to the general public.
void Window::Activate(void) {
	s_allWindows += this;
	MouseHotspotGroup::AddGroup(&m_tabs);
}

/////////////////////////////////////////////////////////////////////////////
//  Deactivate the window.
void Window::Deactivate(void) {
	s_allWindows -= this;
	MouseHotspotGroup::RemoveGroup(&m_tabs);
}

/////////////////////////////////////////////////////////////////////////////
//  Update all of the windows.
void Window::UpdateAll(void) {
	MType m;

	for (m=0; m<s_allWindows.Count(); ++m) {
		s_allWindows[m]->Refresh();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Scroll the window in question.
void Window::CursorUp(void) {
	if (m_panels[m_selection].redraw) {
		m_panels[m_selection].redraw->CursorUp();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Scroll the window in question.
void Window::CursorDown(void) {
	if (m_panels[m_selection].redraw) {
		m_panels[m_selection].redraw->CursorDown();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Scroll the window in question.
void Window::PageUp(void) {
	if (m_panels[m_selection].redraw) {
		m_panels[m_selection].redraw->PageUp();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Scroll the window in question.
void Window::PageDown(void) {
	if (m_panels[m_selection].redraw) {
		m_panels[m_selection].redraw->PageDown();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Generic panel logic.
void Panel::SetDims(u16 x, u16 y, u16 w, u16 h) {
	m_rect.x = x;
	m_rect.y = y;
	m_rect.w = w;
	m_rect.h = h;
}

/////////////////////////////////////////////////////////////////////////////
//  When we click in a panel, the default behavior is to relay the clicks to
// all attached GUI elements.
void Panel::Click(u16 x, u16 y) {
	MType m;

	for (m=0; m<m_guiElements.Count(); ++m) {
		if (m_guiElements[m]->Click(x, y)) {
			break;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Default panel right-clicking (do nothing).
void Panel::RClick(u16 x, u16 y) {
}

/////////////////////////////////////////////////////////////////////////////
//  Add a GUI element to this panel.
void Panel::AddGUIElement(GUIElement * g) {
	m_guiElements += g;
}

/////////////////////////////////////////////////////////////////////////////
//  Redraw the panel.  The root class behavior is to transmit the signal to
// all GUI elements.
void Panel::Redraw(void) {
	MType m;

	for (m=0; m<m_guiElements.Count(); ++m) {
		m_guiElements[m]->Render();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Flag this panel as being dirty.
void Panel::MakeDirty(void) {
	MType m;

	ActiveGraphic::MakeDirty();
	for (m=0; m<m_guiElements.Count(); ++m) {
		m_guiElements[m]->MakeDirty();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Here is the logic for our text windows.
void TextPanel::SetDims(u16 x, u16 y, u16 w, u16 h) {
	Panel::SetDims(x, y, w, h);
	m_content.Make(h / TextGetFontHeight(g_windowFont), __FILE__, __LINE__);
}

/////////////////////////////////////////////////////////////////////////////
//  Take the appropriate action for when we press up.
void TextPanel::CursorUp(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  User presses "down".
void TextPanel::CursorDown(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  Page up.
void TextPanel::PageUp(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  User presses page-down.
void TextPanel::PageDown(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  Go to the indicated line number.
void TextPanel::Goto(u32 line) {
}

/////////////////////////////////////////////////////////////////////////////
//  Okay, this is the redraw function for the text panels.  HEREHEREHERE we
// can eventually specialize our IsDirty() logic and use blitting to scroll.
void TextPanel::Redraw(void) {
	if (IsDirty()) {
		MType m;
		u16 x, y;
		u16 fh = TextGetFontHeight(g_windowFont);
	
		g_monitor.Clear(g_colors[BACKGROUND], &m_rect);
		for (x=m_rect.x, y=m_rect.y, m=0;
			 m<m_content.Count();
			 ++m) {
			TextWrite(g_monitor,
					  g_windowFont,
					  x,
					  y,
					  g_colors[TEXT],
					  m_content[m]);
			y += fh;
		}
		CleanFrame();
	}
	Panel::Redraw();
}

/////////////////////////////////////////////////////////////////////////////
//  By default, a TextPanel will do nothing when you click in it.  So we'll
// just pass it through to the default handler (to get the GUI behavior).
void TextPanel::Click(u16 x, u16 y) {
	Panel::Click(x, y);
}

/////////////////////////////////////////////////////////////////////////////
//  Generate all of the text.
void IListTextPanel::GenerateContent(void) {
	MType m;

	for (m=0;
		 (m + m_lineNumber < m_data.Count()) &&
		 (m < m_content.Count());
		 ++m) {
		m_content[m] = m_data[m + m_lineNumber];
	}
	for (;m < m_content.Count(); ++m) {
		m_content[m].Destroy();
	}
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Move up one line.
void IListTextPanel::CursorUp(void) {
	MType m;

	if (m_lineNumber) {
		for (m=m_content.Count() - 1; m; --m) {
			m_content[m] = m_content[m - 1];
		}
		--m_lineNumber;
		m_content[0] = m_data[m + m_lineNumber];
		MakeDirty();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Move the cursor down one line.
void IListTextPanel::CursorDown(void) {
	MType m;

	if (m_lineNumber + m_content.Count() < m_data.Count()) {
		for (m=0; m < m_content.Count() - 1; ++m) {
			m_content[m] = m_content[m + 1];
		}
		++m_lineNumber;
		m_content[m_content.Count() - 1] =
			m_data[m_content.Count() - 1 + m_lineNumber];
		MakeDirty();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Go up by one page.
void IListTextPanel::PageUp(void) {
	if (m_lineNumber > m_content.Count()) {
		Goto(m_lineNumber - m_content.Count());
	} else {
		Goto(0);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Go down by one page.
void IListTextPanel::PageDown(void) {
	if (m_lineNumber + m_content.Count() < m_data.Count()) {
		Goto(m_lineNumber + m_content.Count());
	} else {
		if (m_content.Count() < m_data.Count()) {
			Goto(m_data.Count() - m_content.Count());
		} else {
			Goto(0);
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Go to the indicated line number.
void IListTextPanel::Goto(u32 line) {
	MType m = (MType)line;

	if (m_content.Count() < m_data.Count()) {
		if (line + m_content.Count() < m_data.Count()) {
			m_lineNumber = line;
		} else {
			m_lineNumber = m_data.Count() - m_content.Count();
		}
	} else {
		m_lineNumber = 0;
	}
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Give the caller access to the strings.
IString & IListTextPanel::operator [](MType m) {
	return m_data[m];
}

/////////////////////////////////////////////////////////////////////////////
//  Add a string to this guy.
void FListTextPanel::InsertString(const FString & f) {
	MType m;

	for (m=0; m<m_content.Count() - 1; ++m) {
		m_content[m] = m_content[m + 1];
	}
	m_content[m] = f;
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  GenerateContent.
void FListTextPanel::GenerateContent(void) {
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  This is just to get things up and running.
void DummyPanel::GenerateContent(void) {
	MType m;

	for (m=0; m<m_content.Count(); ++m) {
		m_content[m] = "* N * Y * I *";
	}
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Construct a GUI element.  Basically, flag it as being dirty.
GUIElement::GUIElement(void) {
	m_visible = TRUE;
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Is this GUI element in need of a graphical refresh?
Bool GUIElement::IsDirty(void) {
	if (m_visible) {
		return ActiveGraphic::IsDirty();
	} else {
		return FALSE;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Hide this GUI element (saves a lot of dynamic insertion/deletion).
void GUIElement::Hide(void) {
	m_visible = FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//  Show this GUI element.
void GUIElement::Show(void) {
	m_visible = TRUE;
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Build the checkbox.
GUICheckBox::GUICheckBox(void) {
	m_spacing = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Specify the name of the checkbox.
void GUICheckBox::SetTitle(const FString & title) {
	m_title = title;
}

/////////////////////////////////////////////////////////////////////////////
//  Set the dimensions (screen position, w and h are constant) of the check-
// box.
void GUICheckBox::SetDims(u16 x, u16 y) {
	m_x = x;
	m_y = y;
}

/////////////////////////////////////////////////////////////////////////////
//  Here, we indicate how many spaces exist between the box and the caption.
// This is to avoid problems with translators not keeping the proper number
// of spaces in the string.
void GUICheckBox::SetSpacing(u8 spaces) {
	m_spacing = spaces;
}

/////////////////////////////////////////////////////////////////////////////
//  Click on the checkbox.  This will check if I've been clicked, and if so,
// takes the appropriate action.
Bool GUICheckBox::Click(u16 x, u16 y) {
	if (m_visible) {
		u16 w = Min(TextGetFontWidth(g_windowFont), TextGetFontHeight(g_windowFont));
	
		if ((x - m_x < w) && (y - m_y < w)) {
			if (QuerySelect()) {
				Deselect();
			} else {
				Select();
			}
			MakeDirty();
			return TRUE;
		} else {
			return FALSE;
		}
	} else {
		return FALSE;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Render the checkbox.
void GUICheckBox::Render(void) {
	if (m_visible && IsDirty()) {
		FString f;
		u8 s = m_spacing;

		f = TEXT_COLOR_GREEN;
		if (QuerySelect()) {
			f += (wchar_t)0x15;
		} else {
			f += (wchar_t)0x14;
		}
		f += (wchar_t)TEXT_COLOR_ORANGE;
		while (s--)
			f += ' ';
		f += m_title;
		TextWrite(g_monitor,
				  g_windowFont,
				  m_x,
				  m_y,
				  g_colors[BORDER],
				  f);
		CleanFrame();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Determine the width of the checkbox.
u16 GUICheckBox::GetWidth(void) const {
	u16 fw = TextGetFontWidth(g_windowFont);

	return fw * (m_title.Length() + m_spacing + 1);
}

/////////////////////////////////////////////////////////////////////////////
//  Determine the height of the checkbox.
u16 GUICheckBox::GetHeight(void) const {
	return TextGetFontHeight(g_windowFont);
}

/////////////////////////////////////////////////////////////////////////////
//  This constructs a radio checkbox.
GUIRadioCheckBox::GUIRadioCheckBox(void) {
	m_selection = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Specify the title of this radio checkbox.
void GUIRadioCheckBox::SetTitle(const FString & title) {
	m_title = title;
}

/////////////////////////////////////////////////////////////////////////////
//  Specify a caption for the indicated item.
void GUIRadioCheckBox::SetCaption(u8 item, const FString & caption) {
	m_captions[item] = caption;
}

/////////////////////////////////////////////////////////////////////////////
//  Specify the dimensions for the GUIRadioCheckBox.
void GUIRadioCheckBox::SetDims(u16 x, u16 y, u8 count) {
	m_x = x;
	m_y = y;
	m_captions.Make(count, __FILE__, __LINE__);
}

/////////////////////////////////////////////////////////////////////////////
//  Click!
Bool GUIRadioCheckBox::Click(u16 x, u16 y) {
	if (m_visible) {
		u16 fw = TextGetFontWidth(g_windowFont);
		u16 fh = TextGetFontHeight(g_windowFont);
		MType m;
		u16 yy = m_y;
	
		for (m=0; m<m_captions.Count(); ++m) {
			yy += fh;
			if ((x - m_x < fw) && (y - yy < fh)) {
				u8 s = m_selection;
	
				if (m != s) {
					Deselect(s);
					Select(m);
					m_selection = QuerySelect();
					MakeDirty();
				}
				return TRUE;
			}
		}
		return FALSE;
	} else {
		return FALSE;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Render the radio checkbox.
void GUIRadioCheckBox::Render(void) {
	if (m_visible && IsDirty()) {
		FString f;
		u16 fh = TextGetFontHeight(g_windowFont);
		u16 yy = m_y;
		MType m;

		TextWrite(g_monitor,
				  g_windowFont,
				  m_x,
				  yy,
				  g_colors[SELECTED],
				  m_title);
		for (m=0; m<m_captions.Count(); ++m) {
			yy += fh;
			if (m == m_selection) {
				f = 0x17;
				f += ' ';
			} else {
				f = 0x16;
				f += ' ';
			}
			f += m_captions[m];
			TextWrite(g_monitor,
					  g_windowFont,
					  m_x,
					  yy,
					  g_colors[TEXT],
					  f);
		}
		CleanFrame();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Determine the width of the radio checkbox.
u16 GUIRadioCheckBox::GetWidth(void) const {
	MType m;
	u16 fw = TextGetFontWidth(g_windowFont);
	u16 retval = fw * m_title.Length();

	for (m=0; m<m_captions.Count(); ++m) {
		u16 u = (m_captions[m].Length() + 2) * fw;

		if (u > retval)
			retval = u;
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Determine the height of the radio checkbox.
u16 GUIRadioCheckBox::GetHeight(void) const {
	u16 fh = TextGetFontHeight(g_windowFont);

	return fh * (m_captions.Count() + 1);
}
